# Copyright 2023 The autobound Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Library for evaluating different types of polynomials."""

import math
import operator
from typing import Callable, Iterator, Sequence, Tuple, TypeVar, Union

from autobound import interval_arithmetic
from autobound import types

Foo = TypeVar('Foo')  # some arbitrary type, like NDArray or Interval
FooLike = TypeVar('FooLike', bound=Foo)  # pytype: disable=invalid-typevar


def eval_polynomial(
    coefficients: Sequence[FooLike],
    z: FooLike,
    inner_product: Callable[[FooLike, FooLike], Foo],
    outer_power: Callable[[FooLike, int], Foo],
    add: Callable[[FooLike, FooLike], Foo] = operator.add,
    additive_identity: Foo = 0,
    multiplicative_identity: Foo = 1) -> Foo:
  """Returns the value of a polynomial at a specific point."""
  running_sum = additive_identity
  z_to_the_i = multiplicative_identity
  for i, coefficient in enumerate(coefficients):
    if i > 0:
      z_to_the_i = outer_power(z, i)
    term = inner_product(coefficient, z_to_the_i)
    running_sum = add(running_sum, term)
  return running_sum


def eval_elementwise_taylor_enclosure(
    enclosure: types.ElementwiseTaylorEnclosureLike,
    x_minus_x0: Union[types.NDArrayLike, types.IntervalLike],
    np_like: types.NumpyLike) -> Union[types.Interval, types.NDArray]:
  """Returns value of an ElementwiseTaylorEnclosure at x-x0."""
  set_arithmetic = interval_arithmetic.IntervalArithmetic(np_like)
  return eval_polynomial(enclosure,
                         set_arithmetic.as_interval_or_ndarray(x_minus_x0),
                         set_arithmetic.multiply,
                         set_arithmetic.power,
                         set_arithmetic.add,
                         np_like.array(0),
                         np_like.array(1))


def eval_taylor_enclosure(
    enclosure: types.TaylorEnclosureLike,
    x_minus_x0: Union[types.NDArrayLike, types.IntervalLike],
    np_like: types.NumpyLike) -> Union[types.Interval, types.NDArray]:
  """Returns value of an TaylorEnclosure at x-x0."""
  set_arithmetic = interval_arithmetic.IntervalArithmetic(np_like)
  inner_product = (
      lambda a, b: set_arithmetic.tensordot(a, b, set_arithmetic.ndim(b)))
  return eval_polynomial(enclosure,
                         set_arithmetic.as_interval_or_ndarray(x_minus_x0),
                         inner_product,
                         set_arithmetic.outer_power,
                         set_arithmetic.add,
                         np_like.array(0),
                         np_like.array(1))


def arbitrary_bilinear(
    a: Sequence[FooLike],
    b: Sequence[FooLike],
    add: Callable[[FooLike, FooLike], Foo] = operator.add,
    additive_identity: Foo = 0,
    term_product_coefficient: Callable[[FooLike, FooLike, int, int], Foo]
    = lambda c0, c1, i, j: c0*c1,
) -> Tuple[Foo, ...]:
  """Applies an arbitrary bilinear operation to two polynomials.

  The arguments a and b give the coefficients of polynomials, defined in terms
  of some inner product <x, y> and some exponentiation operator:

    P_a(z) = sum_{i=0}^{len(a)-1} <a[i], z**i>.

  Similarly, the sequence b represents a polynomial P_b(z).

  Args:
    a: a polynomial (sequence of coefficients)
    b: a polynomial (sequence of coefficients)
    add: a function that returns the sum of two polynomial coefficients
    additive_identity: a addtive identity object
    term_product_coefficient: a callable that, given arguments c0, c1, i, j,
      returns d such that op(<c0, z**i>, <c1, z**j>) = <d, z**(i+j)>, where op
      is the underlying bilinear operation.

  Returns:
    a polynomial Q (tuple of coefficients), such that for any z,
      op(P_a(z), P_b(z)) == Q(z)
    where op is the underlying bilinear operation.
  """
  # By bilinearity,
  # op(sum_i <a[i], z**i>, sum_j <b[j], z**j>)
  # == sum_{ij} op(<a[i], z**i>, <b[j], z**j>)
  # == sum_{ij} <term_product_coefficient(a[i], b[j], i, j), z**(i+j)>.
  output_degree = len(a) + len(b) - 2
  output = [additive_identity] * (output_degree + 1)
  # If a and b have length n, this takes time O(n^2).  If we ever care about
  # large n, we could consider implementing an O(n log n) algorithm using
  # Fourier transforms.
  for i, c0 in enumerate(a):
    for j, c1 in enumerate(b):
      c = term_product_coefficient(c0, c1, i, j)
      output[i+j] = add(output[i+j], c)
  return tuple(output)


def integer_power(
    a: Sequence[FooLike],
    exponent: int,
    add: Callable[[FooLike, FooLike], Foo] = operator.add,
    additive_identity: Foo = 0,
    multiplicative_identity: Foo = 1,
    term_product_coefficient: Callable[[FooLike, FooLike, int, int], Foo]
    = lambda c0, c1, i, j: c0*c1,
    term_power_coefficient: Callable[[FooLike, int, int], Foo]
    = lambda c, i, j: c**j,
    scalar_product: Callable[[int, FooLike], Foo] = operator.mul
) -> Tuple[Foo, ...]:
  """Returns the coefficients of a polynomial raised to a power.

  The arguments a gives the coefficients of a polynomial, defined in terms
  of some inner product <x, y> and some exponentiation operator:

    P_a(z) = sum_{i=0}^{len(a)-1} <a[i], z**i>.

  Let op be some bilinear, associative, and commutative operation.  We define:

    power(a, 0) == multiplicative_identity
    power(a, k) = op(a, power(a, k-1)).

  This code uses the functions provided as arguments to efficiently compute
  the coefficients of the polynomial power(a, exponent).

  When the coefficients of P_a are intervals, this efficient computation
  translates into tighter intervals in the returned coefficients.

  Args:
    a: a polynomial (sequence of coefficients)
    exponent: a non-negative integer exponent
    add: a function that returns the sum of two polynomial coefficients
    additive_identity: a addtive identity object
    multiplicative_identity: a multiplicative identity object
    term_product_coefficient: a callable that, given arguments c0, c1, i, j,
      returns d such that op(<c0, z**i>, <c1, z**j>) = <d, z**(i+j)>, where op
      is the underlying bilinear operation.
    term_power_coefficient: given arguments c, i, and j, returns d such that:
      (c * z**i)**j == d * z**(i*j)
    scalar_product: a callable that, given as arguments a non-negative integer i
      and coefficient c, returns the result of adding c to itself i times.

  Returns:
    the coefficients of the polynomial P_a, raised to the exponent power.
  """
  if exponent < 0:
    raise ValueError(exponent)
  elif exponent == 0:
    return (multiplicative_identity,)
  else:
    # To understand what this code is doing, it is helpful to consider the
    # special case where `a` is a sequence of floats, and all arguments have
    # their default values.  Then, we just need to compute the coefficients
    # of the scalar polynomial:
    #
    #     (a[0] + a[1]*z**1 + ... + a[k-1])**exponent
    #
    # where k = len(a).
    #
    # Using the multinomial theorem, the result is a polynomial whose ith
    # coefficient is:
    #
    #     sum_{p in Partitions(i, exponent, k)}
    #        (exponent choose (p_0, p_1, ..., p_{k-1})) *
    #        Prod_{j=0}^{k-1} a[j]**p_j
    #
    # where Partitions(i, exponent, k) is the set of length-k non-negative
    # integer tuples whose elements sum to `exponent`, and that furthermore
    # satisfy sum_{j=0}^{k-1} j*p_j == i.
    #
    # The code below uses a generalization of this idea that works for an
    # arbitrary commutative and associative bilinear operation (rather than
    # just scalar multiplication).  In the general version, the product
    # series Prod_{j=0}^{k-1} a[j]**p_j is computed via appropriate calls to
    # term_product_coefficient(), term_power_coefficient() and scalar_product().

    def get_coeff(i: int) -> Foo:
      c = additive_identity
      for p in _iter_partitions(i, exponent, len(a)):
        assert sum(p) == exponent
        assert sum(j*p_j for j, p_j in enumerate(p)) == i
        running_product = multiplicative_identity
        running_product_power = 0
        for j, p_j in enumerate(p):
          running_product = term_product_coefficient(
              running_product,
              term_power_coefficient(a[j], j, p_j),
              running_product_power,
              j*p_j
          )
          running_product_power += j*p_j
        assert running_product_power == i
        term = scalar_product(_multinomial_coefficient(p), running_product)
        c = add(c, term)
      return c
    output_degree = (len(a) - 1) * exponent
    return tuple(get_coeff(i) for i in range(1 + output_degree))


def _iter_partitions(
    n: int, m: int, k: int) -> Iterator[Tuple[int, ...]]:
  """Yields length-k tuples with sum m and sum_{j=1}^k (j-1)*i_j == n."""
  if n < 0:
    raise ValueError(n)
  if m < 0:
    raise ValueError(m)
  if k <= 0:
    raise ValueError(k)
  if k == 1:
    if n == 0:
      yield (m,)
  else:
    for z in range(min(m+1, n // (k-1) + 1)):
      for p in _iter_partitions(n - (k-1)*z, m - z, k - 1):
        yield p + (z,)


def _multinomial_coefficient(ks: Sequence[int]) -> int:
  """Returns (n choose (ks[0], ks[1], ...)), where n = sum(ks)."""
  if not ks:
    raise ValueError(ks)
  elif len(ks) == 1:
    return 1
  else:
    return math.comb(sum(ks), ks[0]) * _multinomial_coefficient(ks[1:])



# Copyright 2023 The autobound Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Library for editing JAX expressions (Jaxprs)."""

import itertools
import types
from typing import Any, Tuple, Union

from autobound import graph_editor
import jax


def replace(pattern: jax.core.Jaxpr,
            replacement: jax.core.Jaxpr,
            subject: jax.core.Jaxpr) -> jax.core.Jaxpr:
  """Replace occurrences of a pattern Jaxpr within a subject Jaxpr.

  This method return a `Jaxpr` that is an edited version of `subject`
  in which occurrences of `pattern` have been replaced by `replacement`.
  Occurences of `pattern` must be contiguous sequences of equations in
  `subject`.

  Example usage:

  ```python
  pattern_fun = jnp.exp
  replacement_fun = jnp.log
  subject_fun = lambda x: 2*jnp.exp(x+1)
  to_jaxpr = lambda f: jax.make_jaxpr(f)(0.).jaxpr

  replace(
      to_jaxpr(pattern_fun),
      to_jaxpr(replacement_fun),
      to_jaxpr(subject_fun),
  )  # returns Jaxpr for 2*jnp.log(x+1).
  ```

  Args:
    pattern: a `Jaxpr`
    replacement: a `Jaxpr`
    subject: a `Jaxpr`

  Returns:
    a `Jaxpr` in which occurrences of `pattern` have been replaced by
    `replacement`.
  """
  if len(pattern.invars) != len(replacement.invars):
    raise ValueError()
  if len(pattern.outvars) != len(replacement.outvars):
    raise ValueError()
  if pattern.constvars or replacement.constvars:
    raise NotImplementedError()
  pattern_graph = _jaxpr_to_graph(pattern)
  # When creating the replacement graph, offset the variable counts to keep
  # them unique.
  max_count = max(
      v[1] if v[0] else -1  # pytype: disable=unsupported-operands
      for v in pattern_graph.intermediate_variables()
  )
  replacement_graph = _jaxpr_to_graph(replacement, offset=max_count + 1)
  # In the replacement graph, replace inputs/outputs with those in the pattern
  # graph.
  intermediate_variable_map = {}
  for u, v in itertools.chain(
      zip(replacement_graph.inputs, pattern_graph.inputs),
      zip(replacement_graph.outputs, pattern_graph.outputs),
  ):
    intermediate_variable_map[u] = v
  replacement_operations = [
      e.subs(intermediate_variable_map) for e in replacement_graph.operations
  ]
  hypergraph = graph_editor.replace(
      pattern_graph.operations,
      replacement_operations,
      _jaxpr_to_graph(subject),
      _can_bind,
  )
  return _graph_to_jaxpr(hypergraph)


class _EqnData:

  def __init__(self, primitive, params):
    self.primitive = primitive
    self.params = params

  def __eq__(self, other):
    return (isinstance(other, _EqnData) and
            self.primitive == other.primitive and
            _jaxpr_eqn_params_equiv(self.params, other.params))

  def __hash__(self):
    return hash(self.primitive)


# An intermediate variable is a tuple that either represents a jax.core.Var or
# a jax.core.Literal.
_IntermediateVariable = Union[
    # If the first element of the tuple is True, then the tuple represents
    # a jax.core.Var, and is of the form (True, count, suffix, aval).
    Tuple[bool, int, str, jax.core.AbstractValue],
    # If the first element of the tuple is False, then the tuple represents
    # a jax.core.Literal, and is of the form (False, val, aval).
    Tuple[bool, Any, jax.core.AbstractValue]
]


def _jaxpr_to_graph(jaxpr: jax.core.Jaxpr,
                    offset: int = 0) -> graph_editor.ComputationGraph:
  """Returns a ComputationGraph that represents a Jaxpr.

  Args:
    jaxpr: a `Jaxpr`
    offset: an offset for the indices that appear in intermediate variables.
      This can be used to ensure uniqueness.

  Returns:
    a ComputationGraph that represents `jaxpr`.
  """

  def get_intermediate_variable(
      var_or_literal: Union[jax.core.Var, jax.core.Literal]
  ) -> _IntermediateVariable:
    if isinstance(var_or_literal, jax.core.Var):
      var = var_or_literal
      return (True, var.count + offset, var.suffix, var.aval)
    elif isinstance(var_or_literal, jax.core.Literal):
      literal = var_or_literal
      return (False, literal.val, literal.aval)
    else:
      raise NotImplementedError()

  operations = []
  for eqn in jaxpr.eqns:
    data = _EqnData(eqn.primitive, eqn.params)
    edge = graph_editor.Operation(
        data,
        [get_intermediate_variable(v) for v in eqn.invars],
        [get_intermediate_variable(v) for v in eqn.outvars]
    )
    operations.append(edge)

  data = [get_intermediate_variable(v) for v in jaxpr.constvars]
  return graph_editor.ComputationGraph(
      [get_intermediate_variable(v) for v in jaxpr.invars],
      [get_intermediate_variable(v) for v in jaxpr.outvars],
      operations,
      data=data
  )


def _graph_to_jaxpr(h: graph_editor.ComputationGraph) -> jax.core.Jaxpr:
  """Returns the Jaxpr represented by a ComputationGraph."""
  count_to_var = {}

  def vertex_to_var_or_literal(vertex):
    if vertex[0]:
      _, count, suffix, aval = vertex
      if count not in count_to_var:
        count_to_var[count] = jax.core.Var(count, suffix, aval)
      return count_to_var[count]
    else:
      _, val, aval = vertex
      return jax.core.Literal(val, aval)

  eqns = []
  for edge in h.operations:
    eqn = jax.core.new_jaxpr_eqn(
        invars=[vertex_to_var_or_literal(u) for u in edge.inputs],
        outvars=[vertex_to_var_or_literal(v) for v in edge.outputs],
        primitive=edge.data.primitive,
        params=edge.data.params,
        effects=set()
    )
    eqns.append(eqn)

  invars = [vertex_to_var_or_literal(v) for v in h.inputs]
  outvars = [vertex_to_var_or_literal(v) for v in h.outputs]
  constvars = [vertex_to_var_or_literal(v) for v in h.data]
  return jax.core.Jaxpr(constvars, invars, outvars, eqns)


def _can_bind(u, v):
  if u[0]:
    return v[0]
  else:
    return (not v[0]) and (u[1] == v[1])


# Set of Jaxpr equation params we ignore for matching purposes.
_JAXPR_EQN_PARAMS_TO_IGNORE = frozenset(['weak_type'])


def _jaxpr_eqn_params_equiv(p0, p1) -> bool:
  """Returns whether two Jaxpr equation params dicts are equivalent."""
  if set(p0.keys()) != set(p1.keys()):
    return False
  for k0, v0 in p0.items():
    if k0 in _JAXPR_EQN_PARAMS_TO_IGNORE:
      continue
    v1 = p1[k0]
    if isinstance(v0, jax.core.ClosedJaxpr):
      # TODO(mstreeter): this could incorrectly return True if v0.consts and
      # v1.consts are different.
      if not _same_jaxpr_up_to_variable_renaming(v0.jaxpr, v1.jaxpr,
                                                 ignore_shape=True):
        return False
    elif isinstance(v0, jax.core.Jaxpr):
      if not _same_jaxpr_up_to_variable_renaming(v0, v1, ignore_shape=True):
        return False
    elif isinstance(v0, types.FunctionType):
      if not isinstance(v1, types.FunctionType):
        return False
    elif v0 != v1:
      return False
  return True


def _match_avals(a0, a1, ignore_shape):
  return a0.dtype == a1.dtype and (ignore_shape or (a0.shape == a1.shape))


def _same_jaxpr_up_to_variable_renaming(j0: jax.core.Jaxpr,
                                        j1: jax.core.Jaxpr,
                                        ignore_shape: bool = False) -> bool:
  """Return whether to Jaxprs are identical up to variable renaming."""
  var_map = {}
  def check(v0, v1):
    if isinstance(v0, jax.core.Literal):
      return (isinstance(v1, jax.core.Literal) and v0.val == v1.val and
              v0.aval == v1.aval)
    elif isinstance(v0, jax.core.Var):
      if v0 not in var_map:
        if _match_avals(v0.aval, v1.aval, ignore_shape):
          var_map[v0] = v1
          return True
        else:
          return False
      else:
        return var_map[v0] == v1
    else:
      raise NotImplementedError(v0)

  if (len(j0.constvars) != len(j1.constvars) or
      len(j0.invars) != len(j1.invars) or
      len(j0.outvars) != len(j1.outvars) or
      len(j0.eqns) != len(j1.eqns)):
    return False

  for v0, v1 in itertools.chain(zip(j0.invars, j1.invars),
                                zip(j0.constvars, j1.constvars),
                                zip(j0.outvars, j1.outvars)):
    if not check(v0, v1):
      return False

  for eq0, eq1 in zip(j0.eqns, j1.eqns):
    if eq0.primitive != eq1.primitive:
      return False
    if not _jaxpr_eqn_params_equiv(eq0.params, eq1.params):
      return False
    if (len(eq0.invars) != len(eq1.invars) or
        len(eq0.outvars) != len(eq1.outvars)):
      return False
    for v0, v1 in zip(eq0.invars, eq1.invars):
      if not check(v0, v1):
        return False
    for v0, v1 in zip(eq0.outvars, eq1.outvars):
      if not check(v0, v1):
        return False

  return True


# Copyright 2023 The autobound Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Core types used in this package."""

import abc
from typing import List, NewType, Tuple, Union
import typing_extensions


# An interval is represented as a pair (a, b) of NDArrays of the same shape.
# It represents the set of NDArrays {x: a <= x <= b}, where the inequalities
# are elementwise.
Interval = Tuple['NDArray', 'NDArray']


# A polynomial whose coefficients can be either NDArrays or Intervals.  The
# mathematical formula for the polynomial can be different for different types
# of polynomials (see below).
IntervalPolynomial = Tuple[Union['NDArray', Interval], ...]


# A degree d Taylor enclosure is a polynomial, represented as a tuple of
# coefficients (c0, c1, ..., c_d).  Each coefficient can be either an NDArray
# or an Interval.  In the case where all coefficients are scalar, the
# corresponding polynomial is:
#
#     sum_{i=0}^d c_i (x - x_0)^i.
#
# In the case where x is a rank-k tensor, (x - x_0)^i is replaced by a repeated
# outer product that yields a tensor of rank k*i, and c_i (x - x_0)^i is
# replaced by an inner product that contracts over the last k*i axes of c_i.
#
# Written as Numpy code, the ith term of the sum becomes:
#
#     np.tensordot(c_i, outer_power(x-x_0, i), k*i)
#
# where:
#
#     outer_power(x-x_0, 0) = 1
#     outer_power(x-x_0, i) = np.tensordot(outer_power(x-x_0, i-1), x-x_0, 0).
#
# A given tuple defines a valid Taylor enclosure of a function f at a point
# x0 if f(x) belongs to the interval obtained by evaluating the polynomial
# at x.
#
# Typically the last coefficient of a degree d Taylor enclosure will be an
# Interval, and the first d coefficients will be NDArrays (but this is not
# required).
TaylorEnclosure = NewType('TaylorEnclosure', IntervalPolynomial)


# A degree d _elementwise_ Taylor enclosure of a function f at a point x0
# is a tuple (c_0, c_1, ... c_d), where each coefficient c_i is either an
# NDArray or an Interval, such that
#
#   f(x) in sum_{i=0}^d c_i * (x - x_0)**i for all x.
#
# In general only the last coefficient is an Interval, and the first d-1
# coefficients are NDArrays that match the coefficients of the degree d-1
# Taylor series expansion of f at x0.
ElementwiseTaylorEnclosure = NewType('ElementwiseTaylorEnclosure',
                                     IntervalPolynomial)


# Type for objects that are accepted in place of NDArray arguments.
#
# Sequence[float] is not included in the Union, because for example
# jnp.add([1.], [2.]) does not work.
NDArrayLike = Union['NDArray', int, float]

# Similar types for Intervals, etc.
IntervalLike = Tuple[NDArrayLike, NDArrayLike]
IntervalPolynomialLike = Tuple[Union[NDArrayLike, IntervalLike], ...]
TaylorEnclosureLike = IntervalPolynomialLike
ElementwiseTaylorEnclosureLike = IntervalPolynomialLike


# pylint: disable=multiple-statements


@typing_extensions.runtime_checkable
class NDArray(typing_extensions.Protocol):
  """Protocol for numpy-like n-dimensional arrays.

  This class is used in type annotations when writing functions that
  accept numpy, jax.numpy, or tensorflow.experimental.numpy ndarrays.

  Example usage:

    def mean_squared_error(a: NDArray, x: NDArray, b: NDArray) -> float:
      return float(((a @ x - b)**2).mean())
  """

  #
  # Python magic methods.
  #
  @abc.abstractmethod
  def __abs__(self) -> 'NDArray': pass
  @abc.abstractmethod
  def __add__(self, o: 'NDArrayLike') -> 'NDArray': pass
  @abc.abstractmethod
  def __and__(self, o: 'NDArrayLike') -> 'NDArray': pass
  @abc.abstractmethod
  def __bool__(self) -> 'NDArray': pass
  @abc.abstractmethod
  def __complex__(self) -> complex: pass
  @abc.abstractmethod
  def __eq__(self, o: 'NDArrayLike') -> 'NDArray': pass
  @abc.abstractmethod
  def __float__(self) -> float: pass
  @abc.abstractmethod
  def __floordiv__(self, o: 'NDArrayLike') -> 'NDArray': pass
  @abc.abstractmethod
  def __ge__(self, o: 'NDArrayLike') -> 'NDArray': pass
  @abc.abstractmethod
  def __getitem__(self, key) -> 'NDArray': pass
  @abc.abstractmethod
  def __gt__(self, o: 'NDArrayLike') -> 'NDArray': pass
  @abc.abstractmethod
  def __index__(self) -> int: pass
  @abc.abstractmethod
  def __int__(self) -> int: pass
  @abc.abstractmethod
  def __invert__(self) -> 'NDArray': pass
  @abc.abstractmethod
  def __iter__(self): pass
  @abc.abstractmethod
  def __le__(self, o: 'NDArrayLike') -> 'NDArray': pass
  @abc.abstractmethod
  def __len__(self) -> int: pass
  @abc.abstractmethod
  def __lt__(self, o: 'NDArrayLike') -> 'NDArray': pass
  @abc.abstractmethod
  def __matmul__(self, o: 'NDArrayLike') -> 'NDArray': pass
  @abc.abstractmethod
  def __mod__(self, o: 'NDArrayLike') -> 'NDArray': pass
  @abc.abstractmethod
  def __mul__(self, o: 'NDArrayLike') -> 'NDArray': pass
  @abc.abstractmethod
  def __ne__(self, o: 'NDArrayLike') -> 'NDArray': pass
  @abc.abstractmethod
  def __neg__(self) -> 'NDArray': pass
  @abc.abstractmethod
  def __or__(self, o: 'NDArrayLike') -> 'NDArray': pass
  @abc.abstractmethod
  def __pos__(self) -> 'NDArray': pass
  @abc.abstractmethod
  def __pow__(self, o: 'NDArrayLike') -> 'NDArray': pass
  @abc.abstractmethod
  def __radd__(self, o: 'NDArrayLike') -> 'NDArray': pass
  @abc.abstractmethod
  def __rand__(self, o: 'NDArrayLike') -> 'NDArray': pass
  @abc.abstractmethod
  def __rfloordiv__(self, o: 'NDArrayLike') -> 'NDArray': pass
  @abc.abstractmethod
  def __rmatmul__(self, o: 'NDArrayLike') -> 'NDArray': pass
  @abc.abstractmethod
  def __rmod__(self, o: 'NDArrayLike') -> 'NDArray': pass
  @abc.abstractmethod
  def __rmul__(self, o: 'NDArrayLike') -> 'NDArray': pass
  @abc.abstractmethod
  def __ror__(self, o: 'NDArrayLike') -> 'NDArray': pass
  @abc.abstractmethod
  def __rpow__(self, o: 'NDArrayLike') -> 'NDArray': pass
  @abc.abstractmethod
  def __rsub__(self, o: 'NDArrayLike') -> 'NDArray': pass
  @abc.abstractmethod
  def __rtruediv__(self, o: 'NDArrayLike') -> 'NDArray': pass
  @abc.abstractmethod
  def __rxor__(self, o: 'NDArrayLike') -> 'NDArray': pass
  @abc.abstractmethod
  def __sub__(self, o: 'NDArrayLike') -> 'NDArray': pass
  @abc.abstractmethod
  def __truediv__(self, o: 'NDArrayLike') -> 'NDArray': pass
  @abc.abstractmethod
  def __xor__(self, o: 'NDArrayLike') -> 'NDArray': pass

  #
  # Properties.
  #
  @property
  def dtype(self): raise NotImplementedError(self.__class__)
  @property
  def ndim(self) -> int: raise NotImplementedError(self.__class__)
  @property
  def shape(self): raise NotImplementedError(self.__class__)
  @property
  def size(self): raise NotImplementedError(self.__class__)
  @property
  def T(self): raise NotImplementedError(self.__class__)

  #
  # Public methods.
  #
  @abc.abstractmethod
  def astype(self, *args, **kwargs) -> 'NDArray': pass
  @abc.abstractmethod
  def clip(self, *args, **kwargs) -> 'NDArray': pass
  @abc.abstractmethod
  def max(self, *args, **kwargs) -> 'NDArray': pass
  @abc.abstractmethod
  def mean(self, *args, **kwargs) -> 'NDArray': pass
  @abc.abstractmethod
  def min(self, *args, **kwargs) -> 'NDArray': pass
  @abc.abstractmethod
  def ravel(self, *args) -> 'NDArray': pass
  @abc.abstractmethod
  def reshape(self, *args, **kwargs) -> 'NDArray': pass
  # TODO(mstreeter): list sum() here, once tnp supports it.
  @abc.abstractmethod
  def tolist(self) -> 'NDArray': pass
  @abc.abstractmethod
  def transpose(self, *axes) -> 'NDArray': pass


@typing_extensions.runtime_checkable
class NumpyLike(typing_extensions.Protocol):
  """Numpy-like back end for performing basic operations on n-dim arrays.

  This class is used in type annotations when writing functions that work
  with an arbitrary Numpy-like backend.
  """

  @abc.abstractmethod
  def abs(self, a: NDArrayLike) -> NDArray: pass
  @abc.abstractmethod
  def absolute(self, a: NDArrayLike, *args, **kwargs) -> NDArray: pass
  @abc.abstractmethod
  def add(self, a: NDArrayLike, b: NDArrayLike) -> NDArray: pass
  @abc.abstractmethod
  def all(self, a: NDArrayLike, *args, **kwargs) -> NDArray: pass
  @abc.abstractmethod
  def allclose(self, *args, **kwargs): pass
  @abc.abstractmethod
  def amax(self, *args, **kwargs): pass
  @abc.abstractmethod
  def amin(self, *args, **kwargs): pass
  @abc.abstractmethod
  def angle(self, *args, **kwargs): pass
  @abc.abstractmethod
  def any(self, *args, **kwargs): pass
  @abc.abstractmethod
  def append(self, *args, **kwargs): pass
  @abc.abstractmethod
  def arange(self, *args, **kwargs): pass
  @abc.abstractmethod
  def arccos(self, *args, **kwargs): pass
  @abc.abstractmethod
  def arccosh(self, *args, **kwargs): pass
  @abc.abstractmethod
  def arcsin(self, *args, **kwargs): pass
  @abc.abstractmethod
  def arcsinh(self, *args, **kwargs): pass
  @abc.abstractmethod
  def arctan(self, *args, **kwargs): pass
  @abc.abstractmethod
  def arctan2(self, *args, **kwargs): pass
  @abc.abstractmethod
  def arctanh(self, *args, **kwargs): pass
  @abc.abstractmethod
  def argmax(self, *args, **kwargs): pass
  @abc.abstractmethod
  def argmin(self, *args, **kwargs): pass
  @abc.abstractmethod
  def argsort(self, *args, **kwargs): pass
  @abc.abstractmethod
  def around(self, *args, **kwargs): pass
  @abc.abstractmethod
  def array(self, obj, **kwargs) -> NDArray: pass
  @abc.abstractmethod
  def array_equal(self, *args, **kwargs): pass
  @abc.abstractmethod
  def asarray(self, a) -> NDArray: pass
  @abc.abstractmethod
  def atleast_1d(self, *args, **kwargs): pass
  @abc.abstractmethod
  def atleast_2d(self, *args, **kwargs): pass
  @abc.abstractmethod
  def atleast_3d(self, *args, **kwargs): pass
  @abc.abstractmethod
  def average(self, *args, **kwargs): pass
  @abc.abstractmethod
  def bitwise_and(self, *args, **kwargs): pass
  @abc.abstractmethod
  def bitwise_not(self, *args, **kwargs): pass
  @abc.abstractmethod
  def bitwise_or(self, *args, **kwargs): pass
  @abc.abstractmethod
  def bitwise_xor(self, *args, **kwargs): pass
  @abc.abstractmethod
  def bool_(self, *args, **kwargs): pass
  @abc.abstractmethod
  def broadcast_arrays(self, *args) -> List[NDArray]: pass
  @abc.abstractmethod
  def broadcast_to(self, *args, **kwargs): pass
  @abc.abstractmethod
  def cbrt(self, *args, **kwargs): pass
  @abc.abstractmethod
  def ceil(self, *args, **kwargs): pass
  @abc.abstractmethod
  def clip(self, *args, **kwargs): pass
  @abc.abstractmethod
  def complex128(self, *args, **kwargs): pass
  @abc.abstractmethod
  def complex64(self, *args, **kwargs): pass
  @abc.abstractmethod
  def complex_(self, *args, **kwargs): pass
  @abc.abstractmethod
  def compress(self, *args, **kwargs): pass
  @abc.abstractmethod
  def concatenate(self, *args, **kwargs): pass
  @abc.abstractmethod
  def conj(self, *args, **kwargs): pass
  @abc.abstractmethod
  def conjugate(self, *args, **kwargs): pass
  @abc.abstractmethod
  def copy(self, *args, **kwargs): pass
  @abc.abstractmethod
  def cos(self, *args, **kwargs): pass
  @abc.abstractmethod
  def cosh(self, *args, **kwargs): pass
  @abc.abstractmethod
  def count_nonzero(self, *args, **kwargs): pass
  @abc.abstractmethod
  def cross(self, *args, **kwargs): pass
  @abc.abstractmethod
  def cumprod(self, *args, **kwargs): pass
  @abc.abstractmethod
  def cumsum(self, *args, **kwargs): pass
  @abc.abstractmethod
  def deg2rad(self, *args, **kwargs): pass
  @abc.abstractmethod
  def diag(self, *args, **kwargs) -> NDArray: pass
  @abc.abstractmethod
  def diag_indices(self, *args, **kwargs): pass
  @abc.abstractmethod
  def diagflat(self, *args, **kwargs): pass
  @abc.abstractmethod
  def diagonal(self, *args, **kwargs): pass
  @abc.abstractmethod
  def diff(self, *args, **kwargs): pass
  @abc.abstractmethod
  def divide(self, a: NDArrayLike, b: NDArrayLike) -> NDArray: pass
  @abc.abstractmethod
  def divmod(self, *args, **kwargs): pass
  @abc.abstractmethod
  def dot(self, *args, **kwargs): pass
  @abc.abstractmethod
  def dsplit(self, *args, **kwargs): pass
  @abc.abstractmethod
  def dstack(self, *args, **kwargs): pass
  @abc.abstractmethod
  def einsum(self, *args, **kwargs): pass
  @abc.abstractmethod
  def empty(self, *args, **kwargs): pass
  @abc.abstractmethod
  def empty_like(self, *args, **kwargs): pass
  @abc.abstractmethod
  def equal(self, x1: NDArrayLike, x2: NDArrayLike, **kwargs) -> NDArray: pass
  @abc.abstractmethod
  def exp(self, x: NDArrayLike) -> NDArray: pass
  @abc.abstractmethod
  def exp2(self, *args, **kwargs): pass
  @abc.abstractmethod
  def expand_dims(self, *args, **kwargs): pass
  @abc.abstractmethod
  def expm1(self, *args, **kwargs): pass
  @abc.abstractmethod
  def eye(self, n: int, *args, **kwargs) -> NDArray: pass
  @abc.abstractmethod
  def fabs(self, *args, **kwargs): pass
  @abc.abstractmethod
  def finfo(self, *args, **kwargs): pass
  @abc.abstractmethod
  def fix(self, *args, **kwargs): pass
  @abc.abstractmethod
  def flip(self, *args, **kwargs): pass
  @abc.abstractmethod
  def fliplr(self, *args, **kwargs): pass
  @abc.abstractmethod
  def flipud(self, *args, **kwargs): pass
  @abc.abstractmethod
  def float16(self, *args, **kwargs): pass
  @abc.abstractmethod
  def float32(self, *args, **kwargs): pass
  @abc.abstractmethod
  def float64(self, *args, **kwargs): pass
  @abc.abstractmethod
  def float_(self, *args, **kwargs): pass
  @abc.abstractmethod
  def float_power(self, *args, **kwargs): pass
  @abc.abstractmethod
  def floor(self, *args, **kwargs): pass
  @abc.abstractmethod
  def floor_divide(self, *args, **kwargs): pass
  @abc.abstractmethod
  def full(self, *args, **kwargs): pass
  @abc.abstractmethod
  def full_like(self, *args, **kwargs): pass
  @abc.abstractmethod
  def gcd(self, *args, **kwargs): pass
  @abc.abstractmethod
  def geomspace(self, *args, **kwargs): pass
  @abc.abstractmethod
  def greater(self, x1: NDArrayLike, x2: NDArrayLike, **kwargs) -> NDArray:
    pass
  @abc.abstractmethod
  def greater_equal(self, x1: NDArrayLike, x2: NDArrayLike,
                    **kwargs) -> NDArray: pass
  @abc.abstractmethod
  def heaviside(self, *args, **kwargs): pass
  @abc.abstractmethod
  def hsplit(self, *args, **kwargs): pass
  @abc.abstractmethod
  def hstack(self, *args, **kwargs): pass
  @abc.abstractmethod
  def hypot(self, *args, **kwargs): pass
  @abc.abstractmethod
  def identity(self, *args, **kwargs): pass
  @abc.abstractmethod
  def iinfo(self, *args, **kwargs): pass
  @abc.abstractmethod
  def imag(self, *args, **kwargs): pass
  @abc.abstractmethod
  def inexact(self, *args, **kwargs): pass
  @abc.abstractmethod
  def inner(self, *args, **kwargs): pass
  @abc.abstractmethod
  def int16(self, *args, **kwargs): pass
  @abc.abstractmethod
  def int32(self, *args, **kwargs): pass
  @abc.abstractmethod
  def int64(self, *args, **kwargs): pass
  @abc.abstractmethod
  def int8(self, *args, **kwargs): pass
  @abc.abstractmethod
  def int_(self, *args, **kwargs): pass
  @abc.abstractmethod
  def isclose(self, *args, **kwargs): pass
  @abc.abstractmethod
  def iscomplex(self, *args, **kwargs): pass
  @abc.abstractmethod
  def iscomplexobj(self, *args, **kwargs): pass
  @abc.abstractmethod
  def isfinite(self, *args, **kwargs): pass
  @abc.abstractmethod
  def isinf(self, *args, **kwargs): pass
  @abc.abstractmethod
  def isnan(self, *args, **kwargs): pass
  @abc.abstractmethod
  def isneginf(self, *args, **kwargs): pass
  @abc.abstractmethod
  def isposinf(self, *args, **kwargs): pass
  @abc.abstractmethod
  def isreal(self, *args, **kwargs): pass
  @abc.abstractmethod
  def isrealobj(self, *args, **kwargs): pass
  @abc.abstractmethod
  def isscalar(self, *args, **kwargs): pass
  @abc.abstractmethod
  def issubdtype(self, *args, **kwargs): pass
  @abc.abstractmethod
  def ix_(self, *args, **kwargs): pass
  @abc.abstractmethod
  def kron(self, *args, **kwargs): pass
  @abc.abstractmethod
  def lcm(self, *args, **kwargs): pass
  @abc.abstractmethod
  def less(self, x1: NDArrayLike, x2: NDArrayLike, **kwargs) -> NDArray: pass
  @abc.abstractmethod
  def less_equal(self, x1: NDArrayLike, x2: NDArrayLike, **kwargs) -> NDArray:
    pass
  @abc.abstractmethod
  def linspace(self, *args, **kwargs): pass
  @abc.abstractmethod
  def log(self, x: NDArrayLike) -> NDArray: pass
  @abc.abstractmethod
  def log1p(self, x: NDArrayLike) -> NDArray: pass
  def log10(self, *args, **kwargs): pass
  @abc.abstractmethod
  def log2(self, *args, **kwargs): pass
  @abc.abstractmethod
  def logaddexp(self, *args, **kwargs): pass
  @abc.abstractmethod
  def logaddexp2(self, *args, **kwargs): pass
  @abc.abstractmethod
  def logical_and(self, a: NDArrayLike, b: NDArrayLike) -> NDArray: pass
  @abc.abstractmethod
  def logical_not(self, *args, **kwargs): pass
  @abc.abstractmethod
  def logical_or(self, *args, **kwargs): pass
  @abc.abstractmethod
  def logical_xor(self, *args, **kwargs): pass
  @abc.abstractmethod
  def logspace(self, *args, **kwargs): pass
  @abc.abstractmethod
  def matmul(self, *args, **kwargs): pass
  @abc.abstractmethod
  def max(self, *args, **kwargs): pass
  @abc.abstractmethod
  def maximum(self, a: NDArrayLike, b: NDArrayLike) -> NDArray: pass
  @abc.abstractmethod
  def mean(self, *args, **kwargs): pass
  @abc.abstractmethod
  def meshgrid(self, *args, **kwargs): pass
  @abc.abstractmethod
  def min(self, *args, **kwargs): pass
  @abc.abstractmethod
  def minimum(self, a: NDArrayLike, b: NDArrayLike) -> NDArray: pass
  @abc.abstractmethod
  def mod(self, *args, **kwargs): pass
  @abc.abstractmethod
  def moveaxis(self, *args, **kwargs): pass
  @abc.abstractmethod
  def multiply(self, a: NDArrayLike, b: NDArrayLike) -> NDArray: pass
  @abc.abstractmethod
  def nanmean(self, *args, **kwargs): pass
  @abc.abstractmethod
  def nanprod(self, *args, **kwargs): pass
  @abc.abstractmethod
  def nansum(self, *args, **kwargs): pass
  @abc.abstractmethod
  def ndarray(self, *args, **kwargs): pass
  @abc.abstractmethod
  def ndim(self, a: NDArrayLike) -> int: pass
  @abc.abstractmethod
  def negative(self, a: NDArrayLike) -> NDArray: pass
  @abc.abstractmethod
  def nextafter(self, *args, **kwargs): pass
  @abc.abstractmethod
  def nonzero(self, *args, **kwargs): pass
  @abc.abstractmethod
  def not_equal(self, *args, **kwargs): pass
  @abc.abstractmethod
  def object_(self, *args, **kwargs): pass
  @abc.abstractmethod
  def ones(self, shape, **kwargs) -> NDArray: pass
  @abc.abstractmethod
  def ones_like(self, a) -> NDArray: pass
  @abc.abstractmethod
  def outer(self, *args, **kwargs): pass
  @abc.abstractmethod
  def pad(self, *args, **kwargs): pass
  @abc.abstractmethod
  def polyval(self, *args, **kwargs): pass
  @abc.abstractmethod
  def positive(self, *args, **kwargs): pass
  @abc.abstractmethod
  def power(self, a, exponent) -> NDArray: pass
  @abc.abstractmethod
  def prod(self, *args, **kwargs): pass
  @abc.abstractmethod
  def promote_types(self, *args, **kwargs): pass
  @abc.abstractmethod
  def ptp(self, *args, **kwargs): pass
  @abc.abstractmethod
  def rad2deg(self, *args, **kwargs): pass
  @abc.abstractmethod
  def ravel(self, *args, **kwargs): pass
  @abc.abstractmethod
  def real(self, *args, **kwargs): pass
  @abc.abstractmethod
  def reciprocal(self, *args, **kwargs): pass
  @abc.abstractmethod
  def remainder(self, *args, **kwargs): pass
  @abc.abstractmethod
  def repeat(self, *args, **kwargs): pass
  @abc.abstractmethod
  def reshape(self, a: NDArrayLike, shape) -> NDArray: pass
  @abc.abstractmethod
  def result_type(self, *args, **kwargs): pass
  @abc.abstractmethod
  def roll(self, *args, **kwargs): pass
  @abc.abstractmethod
  def rot90(self, *args, **kwargs): pass
  @abc.abstractmethod
  def round(self, *args, **kwargs): pass
  @abc.abstractmethod
  def select(self, *args, **kwargs): pass
  @abc.abstractmethod
  def shape(self, a: NDArrayLike): pass
  @abc.abstractmethod
  def sign(self, a: NDArrayLike) -> NDArray: pass
  @abc.abstractmethod
  def signbit(self, *args, **kwargs): pass
  @abc.abstractmethod
  def sin(self, *args, **kwargs): pass
  @abc.abstractmethod
  def sinc(self, *args, **kwargs): pass
  @abc.abstractmethod
  def sinh(self, *args, **kwargs): pass
  @abc.abstractmethod
  def size(self, *args, **kwargs): pass
  @abc.abstractmethod
  def sort(self, *args, **kwargs): pass
  @abc.abstractmethod
  def split(self, *args, **kwargs): pass
  @abc.abstractmethod
  def sqrt(self, *args, **kwargs): pass
  @abc.abstractmethod
  def square(self, *args, **kwargs): pass
  @abc.abstractmethod
  def squeeze(self, *args, **kwargs): pass
  @abc.abstractmethod
  def stack(self, *args, **kwargs): pass
  @abc.abstractmethod
  def std(self, *args, **kwargs): pass
  @abc.abstractmethod
  def subtract(self, a: NDArrayLike, b: NDArrayLike) -> NDArray: pass
  @abc.abstractmethod
  def sum(self, a: NDArrayLike, **kwargs) -> NDArray: pass
  @abc.abstractmethod
  def swapaxes(self, *args, **kwargs): pass
  @abc.abstractmethod
  def take(self, *args, **kwargs): pass
  @abc.abstractmethod
  def take_along_axis(self, *args, **kwargs): pass
  @abc.abstractmethod
  def tan(self, *args, **kwargs): pass
  @abc.abstractmethod
  def tanh(self, *args, **kwargs): pass
  @abc.abstractmethod
  def tensordot(self, a: NDArrayLike, b: NDArrayLike, axes) -> NDArray: pass
  @abc.abstractmethod
  def tile(self, *args, **kwargs): pass
  @abc.abstractmethod
  def trace(self, *args, **kwargs): pass
  @abc.abstractmethod
  def transpose(self, *args, **kwargs): pass
  @abc.abstractmethod
  def tri(self, *args, **kwargs): pass
  @abc.abstractmethod
  def tril(self, *args, **kwargs): pass
  @abc.abstractmethod
  def triu(self, *args, **kwargs): pass
  @abc.abstractmethod
  def true_divide(self, *args, **kwargs): pass
  @abc.abstractmethod
  def uint16(self, *args, **kwargs): pass
  @abc.abstractmethod
  def uint32(self, *args, **kwargs): pass
  @abc.abstractmethod
  def uint64(self, *args, **kwargs): pass
  @abc.abstractmethod
  def uint8(self, *args, **kwargs): pass
  @abc.abstractmethod
  def vander(self, *args, **kwargs): pass
  @abc.abstractmethod
  def var(self, *args, **kwargs): pass
  @abc.abstractmethod
  def vdot(self, *args, **kwargs): pass
  @abc.abstractmethod
  def vsplit(self, *args, **kwargs): pass
  @abc.abstractmethod
  def vstack(self, *args, **kwargs): pass
  @abc.abstractmethod
  def where(self, condition: NDArrayLike, x, y) -> NDArray: pass
  @abc.abstractmethod
  def zeros(self, shape, **kwargs) -> NDArray: pass
  @abc.abstractmethod
  def zeros_like(self, a) -> NDArray: pass

  #
  # Properties.
  #
  @property
  def e(self): raise NotImplementedError()
  @property
  def inf(self): raise NotImplementedError()
  @property
  def newaxis(self): raise NotImplementedError()
  @property
  def pi(self): raise NotImplementedError()



# Copyright 2023 The autobound Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Code for performing arithmetic on Taylor enclosures."""

import functools
from typing import Callable, Optional, Union

from autobound import interval_arithmetic
from autobound import polynomials
from autobound import primitive_enclosures
# pylint: disable=g-multiple-import
from autobound.types import (
    ElementwiseTaylorEnclosure, ElementwiseTaylorEnclosureLike, Interval,
    IntervalLike, NDArray, NDArrayLike, NumpyLike, TaylorEnclosure,
    TaylorEnclosureLike)


class TaylorEnclosureArithmetic:
  """Taylor enclosure arithmetic via a numpy-like back end.

  Given a specified trust region, maximum polynomial degree D, and NumpyLike
  back end, this class allows you to perform arithmetic operations on
  TaylorEnclosures, and get back TaylorEnclosures of degree <= D.

  What does it mean to perform arithmetic operations on Taylor enclosures?
  Let op(arg_0(x), arg_1(x)) be some binary operation (such as addition or
  multiplication).  Let enclosure_i be a Taylor enclosure for arg_i:

      arg_i(x) in enclosure_i(x)  for x-x0 in trust_region.

  Then, op(enclosure_0, enclosure_1) is a Taylor enclosure f, such that:

      op(arg_0(x), arg_1(x)) in f(x)  for x-x0 in trust_region.

  Example usage:
    import jax.numpy as jnp
    trust_region = (jnp.zeros((3,)), jnp.array([1, 2, 3]))
    arithmetic = TaylorEnclosureArithmetic(2, trust_region, jnp)
    arithmetic.power(
        TaylorEnclosure((0, 1),
        4
    )  # ==> quadratic TaylorEnclosure of x**4, valid for x in trust_region
  """

  def __init__(self,
               max_degree: int,
               trust_region: IntervalLike,
               np_like: NumpyLike):
    """Initializer.

    Args:
      max_degree: the maximum degree polynomial output by any of the functions
        in this class
      trust_region: an interval that contains x-x0 (not x!).
      np_like: a NumpyLike Module
    """
    if np_like.shape(trust_region[0]) != np_like.shape(trust_region[1]):
      raise ValueError(trust_region)
    if not isinstance(max_degree, int):
      raise ValueError(max_degree)

    self.max_degree = max_degree
    self.set_arithmetic = interval_arithmetic.IntervalArithmetic(np_like)
    self.trust_region = self.set_arithmetic.as_interval(trust_region)
    self.np_like = np_like

  def add(self,
          a: TaylorEnclosureLike,
          b: TaylorEnclosureLike) -> TaylorEnclosure:
    """Returns the (possibly-truncated) sum of two TaylorEnclosures."""
    a = as_taylor_enclosure(a, self.np_like)
    b = as_taylor_enclosure(b, self.np_like)

    def get_coefficient(i):
      if i < len(a) and i < len(b):
        return self.set_arithmetic.add(a[i], b[i])
      elif i < len(a):
        return a[i]
      else:
        return b[i]

    sum_ab = tuple(get_coefficient(i) for i in range(max(len(a), len(b))))
    return self._truncate_if_necessary(sum_ab)

  def arbitrary_bilinear(
      self,
      a: TaylorEnclosureLike,
      b: TaylorEnclosureLike,
      pairwise_batched_bilinear: Callable[[NDArray, NDArray, int, int], NDArray]
  ) -> TaylorEnclosure:
    """Applies an arbitrary bilinear operation to two TaylorEnclosures.

    Args:
      a: a TaylorEnclosure
      b: a TaylorEnclosure
      pairwise_batched_bilinear: a callable that takes parameters (u, v, p, q),
        and returns the result of applying some underlying operation 'bilinear'
        to various pairs of arguments, where the first argument is indexed by
        the last p dimensions of u, and the second argument is indexed by the
        last q dimensions of v.

        In the special case p=q=1,

            pairwise_batched_bilinear(u, v, 1, 1)[..., i, j]

                == bilinear(u[..., i], v[..., j])  for all i, j.

        In the general case, for all tuples I = (i_1, i_2, ..., i_p), and
        J = (j_1, j_2, ..., j_q),

            pairwise_batched_bilinear(u, v, p, q)[(...,) + I + J]

                == bilinear(u[(...,) + I], v[(...,) + J]) .

    Returns:
      a TaylorEnclosure c, such that for x-x0 in self.trust_region:

        arg_0 in a(x) and arg_1 in b(x) ==> bilinear(arg0, arg1) in c(x)

      where 'bilinear' is pairwise_batched_bilinear's underlying operation.
    """
    # Our goal is to compute a Taylor enclosure for bilinear(a, b), where:
    #
    #    a = sum_i <a[i], (x-x0)^i>
    #    b = sum_j <b[j], (x-x0)^j>
    #
    # Here, we use ^ to denote self.set_arithmetic.outer_power, and we use
    # <u, v> to denote self.set_arithmetic.tensordot(u, v, v.ndim).
    #
    # Using bilinearity, we can show:
    #
    #    bilinear(a, b)
    #        == sum_{i, j} bilinear(<a[i], (x-x0)^i>, <b[i], (x-x0)^j>).
    #
    # Thus, it suffices to compute a Taylor enclosure for each term, and then
    # sum those up.
    #
    # How do we compute a Taylor enclosure for each term?  It can be shown that:
    #
    #     bilinear(<a[i], (x-x0)^i>, <b[j], (x-x0)^j>)
    #     == bilinear(pairwise_batched_bilinear(a[i], b[j], i*x_ndim, j*x_ndim),
    #                 (x-x0)^(i+j)) .
    #
    # If i + j <= self.max_degree, we use simply use this formula to return
    # the coefficient for term (i, j).  Otherwise, we make use of the trust
    # region to enclose the left hand side in terms of (x-x0)^self.max_degree.

    x_ndim = self.np_like.ndim(self.trust_region[0])

    def get_term_enclosure_coefficient(
        i: int, j: int) -> Union[NDArray, Interval]:
      r"""Returns coefficient of enclosure for term coming from (a[i], b[j]).

      Args:
        i: an index into a
        j: an index into b

      Returns:
        an NDArray or Interval c, such that:

            bilinear(inner(a[i], z^i), inner(b[j], z^j)) \subseteq inner(c, z^k)

        where k = min(i+j, self.max_degree).
      """
      if i + j <= self.max_degree:
        c0, c1, c0_power, c1_power = (a[i], b[j], i, j)
      else:
        excess_degree = i + j - self.max_degree
        # Distribute the excess degree between the two factors.
        excess_i = min(i, excess_degree)
        excess_j = excess_degree - excess_i
        assert (i - excess_i) + (j - excess_j) == self.max_degree
        c0 = self.set_arithmetic.tensordot(
            a[i],
            self.set_arithmetic.outer_power(self.trust_region, excess_i),
            excess_i * x_ndim
        )
        if excess_j > 0:
          c1 = self.set_arithmetic.tensordot(
              b[j],
              self.set_arithmetic.outer_power(self.trust_region, excess_j),
              excess_j * x_ndim
          )
        else:
          c1 = b[j]
        c0_power = i - excess_i
        c1_power = j - excess_j

      def bilinear(y, z):
        return pairwise_batched_bilinear(y, z, c0_power*x_ndim,
                                         c1_power*x_ndim)

      return self.set_arithmetic.arbitrary_bilinear(c0, c1, bilinear)

    # Sum up the coefficients from each term.
    output_degree = min(self.max_degree, len(a) + len(b) - 2)
    product_coefficients = [self.np_like.asarray(0)] * (output_degree + 1)
    for i in range(len(a)):
      for j in range(len(b)):
        term_coefficient = get_term_enclosure_coefficient(i, j)
        term_degree = min(i+j, self.max_degree)
        product_coefficients[term_degree] = self.set_arithmetic.add(
            product_coefficients[term_degree],
            term_coefficient,
        )

    return TaylorEnclosure(tuple(product_coefficients))

  def compose_enclosures(
      self,
      elementwise_enclosure: ElementwiseTaylorEnclosureLike,
      arg_enclosure: TaylorEnclosureLike) -> TaylorEnclosure:
    """Returns composition of two enclosures."""
    # TODO(mstreeter): we can potentially do something more efficient than
    # computing each term separately and summing the results.
    if not elementwise_enclosure:
      raise ValueError()
    arg_diff_enclosure = (0,) + arg_enclosure[1:]
    output = TaylorEnclosure((self.np_like.array(0),))

    def interval_left_broadcasting_multiply(a, b):
      bilinear = functools.partial(_left_broadcasting_multiply,
                                   np_like=self.np_like)
      return self.set_arithmetic.arbitrary_bilinear(a, b, bilinear,
                                                    assume_product=True)

    for p, coefficient in enumerate(elementwise_enclosure):
      if p == 0:
        term = (coefficient,)
      else:
        poly = self.power(arg_diff_enclosure, p)
        term = tuple(
            # The special-casing when i < p ensures that the TaylorEnclosure
            # returned by this function will not contain trivial intervals
            # the form (x, x).
            #
            # The issue is that the first p elements of 'poly' are guaranteed
            # to be 0, but the expression below can express this as the interval
            # (0, 0).
            0 if i < p else interval_left_broadcasting_multiply(coefficient, t)
            for i, t in enumerate(poly)
        )
      output = self.add(output, term)
    assert output is not None
    return output

  def divide(self,
             a: TaylorEnclosureLike,
             b: TaylorEnclosureLike) -> TaylorEnclosure:
    return self.multiply(a, self.power(b, -1))

  def get_elementwise_fun(
      self,
      get_elementwise_enclosure: Callable[
          [NDArray, Interval, int, NumpyLike],
          ElementwiseTaylorEnclosure
      ]):
    """Returns elementwise function that inputs/output TaylorEnclosures."""
    def fun(
        arg_enclosure: TaylorEnclosureLike,
        arg_trust_region: Optional[Union[NDArrayLike, IntervalLike]] = None
    ) -> TaylorEnclosure:
      if arg_trust_region is None:
        # If arg_trust_region is not provided derive it from arg_enclosure.
        degree_0_enclosure = enclose_enclosure(arg_enclosure, self.trust_region,
                                               0, self.np_like)
        arg_trust_region = degree_0_enclosure[0]
      arg_trust_region = self.set_arithmetic.as_interval_or_ndarray(
          arg_trust_region)
      if not isinstance(arg_trust_region, tuple):
        arg_trust_region = (arg_trust_region, arg_trust_region)
      x0 = arg_enclosure[0]
      if isinstance(x0, tuple):
        assert self.max_degree == 0
        assert len(x0) == 2
        x0 = x0[0]
      elementwise_enclosure = get_elementwise_enclosure(
          x0,
          arg_trust_region,
          self.max_degree,
          self.np_like)
      return self.compose_enclosures(elementwise_enclosure, arg_enclosure)
    return fun

  def multiply(self,
               a: TaylorEnclosureLike,
               b: TaylorEnclosureLike) -> TaylorEnclosure:
    """Returns elementwise product of two TaylorEnclosures."""
    self._validate_taylor_enclosure(a)
    self._validate_taylor_enclosure(b)
    term_product_coefficient = functools.partial(
        _elementwise_term_product_coefficient,
        x_ndim=self.np_like.ndim(self.trust_region[0]),
        np_like=self.np_like)
    product = polynomials.arbitrary_bilinear(
        a,
        b,
        self.set_arithmetic.add,
        self.np_like.asarray(0),
        term_product_coefficient
    )
    return self._truncate_if_necessary(product)

  def negative(self, a: TaylorEnclosureLike) -> TaylorEnclosure:
    return self._truncate_if_necessary(
        TaylorEnclosure(tuple(self.set_arithmetic.negative(c) for c in a)))

  def power(self, a: TaylorEnclosureLike, p: float) -> TaylorEnclosure:
    """Returns a TaylorEnclosure for a**p, of degree <= self.max_degree."""
    self._validate_taylor_enclosure(a)
    if p >= 0 and p == int(p):
      x_ndim = self.np_like.ndim(self.trust_region[0])
      term_product_coefficient = functools.partial(
          _elementwise_term_product_coefficient, x_ndim=x_ndim,
          np_like=self.np_like)
      term_power_coefficient = functools.partial(
          _elementwise_term_power_coefficient, x_ndim=x_ndim,
          np_like=self.np_like)
      multiplicative_identity = self.np_like.ones_like(self.trust_region[0])
      result = polynomials.integer_power(  # pytype: disable=wrong-arg-types
          a,
          p,
          self.set_arithmetic.add,
          self.np_like.asarray(0),
          multiplicative_identity,
          term_product_coefficient,
          term_power_coefficient,
          self.set_arithmetic.multiply
      )
      return self._truncate_if_necessary(result)
    else:
      get_elementwise_enclosure = functools.partial(
          primitive_enclosures.pow_enclosure, p)
      return self.get_elementwise_fun(get_elementwise_enclosure)(a)

  def subtract(self,
               a: TaylorEnclosureLike,
               b: TaylorEnclosureLike) -> TaylorEnclosure:
    return self.add(a, self.negative(b))

  def _truncate_if_necessary(self, a: TaylorEnclosureLike) -> TaylorEnclosure:
    return enclose_enclosure(a, self.trust_region, self.max_degree,
                             self.np_like)

  def _validate_taylor_enclosure(self, a: TaylorEnclosureLike):
    x_shape = self.set_arithmetic.shape(self.trust_region)
    for i, coeff in enumerate(a):
      s = self.set_arithmetic.shape(coeff)
      if s[len(s)-i*len(x_shape):] != i*x_shape:
        raise ValueError(x_shape, i, s, coeff, a)


def as_taylor_enclosure(a: TaylorEnclosureLike,
                        np_like: NumpyLike) -> TaylorEnclosure:
  set_arithmetic = interval_arithmetic.IntervalArithmetic(np_like)
  return TaylorEnclosure(
      tuple(set_arithmetic.as_interval_or_ndarray(c) for c in a))


def enclose_enclosure(
    enclosure: TaylorEnclosureLike,
    trust_region: IntervalLike,
    max_degree: int,
    np_like: NumpyLike,
) -> TaylorEnclosure:
  """Returns a (possibly) lower-degree enclosure of a given TaylorEnclosure."""
  set_arithmetic = interval_arithmetic.IntervalArithmetic(np_like)
  trust_region = set_arithmetic.as_interval(trust_region)
  enclosure = as_taylor_enclosure(enclosure, np_like)
  orig_degree = len(enclosure) - 1
  if orig_degree <= max_degree:
    return enclosure
  else:
    new_final_coefficient = polynomials.eval_taylor_enclosure(
        enclosure[max_degree:], trust_region, set_arithmetic.np_like)
    return TaylorEnclosure(enclosure[:max_degree] + (new_final_coefficient,))


def expand_multiple_dims(a: NDArray, n: int, axis=None) -> NDArray:
  """Like expand_dims, but adds n dims rather than just 1."""
  if axis is None:
    axis = a.ndim
  colon = slice(None, None, None)
  return a[(colon,) * axis + (None,) * n + (...,)]


def map_over_enclosure(
    a: TaylorEnclosure,
    fun: Callable[[NDArray], NDArray]) -> TaylorEnclosure:
  """Apply a function to each NDArray in a TaylorEnclosure."""
  return TaylorEnclosure(
      tuple(tuple(map(fun, c)) if isinstance(c, tuple) else fun(c) for c in a))


def _elementwise_term_power_coefficient(
    c: Union[NDArrayLike, IntervalLike],
    i: int,
    exponent: int,
    x_ndim: int,
    np_like: NumpyLike) -> Union[NDArray, Interval]:
  """Returns d such that <c, z**i>^exponent == <d, z**(i*exponent)>.

  Args:
    c: a coefficient
    i: a non-negative integer
    exponent: a non-negative integer
    x_ndim: the number of dimensions in the independent variable
    np_like: a Numpy-like backend

  Returns:
    an NDArray or Interval d, such that
        <c, z**i>^exponent == <d, z**(i*exponent)>.
    where ** denotes outer product, and ^ denotes elementwise exponentiation.
  """
  set_arithmetic = interval_arithmetic.IntervalArithmetic(np_like)
  batch_dims = set_arithmetic.ndim(c) - i*x_ndim
  if batch_dims < 0:
    raise ValueError((set_arithmetic.ndim(c), i, x_ndim))
  return set_arithmetic.outer_power(c, exponent, batch_dims)


def _elementwise_term_product_coefficient(
    c0: Union[NDArrayLike, IntervalLike],
    c1: Union[NDArrayLike, IntervalLike],
    i: int,
    j: int,
    x_ndim: int,
    np_like: NumpyLike) -> Union[NDArray, Interval]:
  """Returns d such that <c0, z**i> * <c1, z**j> == <d, z**(i+j)>."""
  def product(u, v):
    return _pairwise_batched_multiply(u, v, i*x_ndim, j*x_ndim, np_like)
  set_arithmetic = interval_arithmetic.IntervalArithmetic(np_like)
  return set_arithmetic.arbitrary_bilinear(c0, c1, product, assume_product=True)


def _pairwise_batched_multiply(
    u: NDArrayLike,
    v: NDArrayLike,
    p: int,
    q: int,
    np_like: NumpyLike) -> NDArray:
  """Batched version of multiply, for use as input to arbitrary_bilinear().

  See the docstring for TaylorEnclosureArithmetic.arbitrary_bilinear for
  context.

  Args:
    u: an NDArray of dimension at least p
    v: an NDArray of dimension at least q
    p: a non-negative integer
    q: a non-negative integer
    np_like: a NumpyLike back end

  Returns:
    an NDArray 'output', such that for every pair of tuples
    I = (i_1, i_2, ..., i_p) and J = (j_1, j_2, ..., j_q),

      output[(...,) + I + J] = u[(...,) + I] * v[(...,) + J] .
  """
  u = np_like.asarray(u)
  v = np_like.asarray(v)
  return expand_multiple_dims(u, q) * expand_multiple_dims(v, p, v.ndim-q)


def _left_broadcasting_multiply(a: NDArrayLike, b: NDArrayLike,
                                np_like: NumpyLike) -> NDArray:
  """Multiplies a and b, broadcasting over leftmost dimensions."""
  a = np_like.asarray(a)
  b = np_like.asarray(b)
  if a.ndim > b.ndim:
    raise NotImplementedError()
  return expand_multiple_dims(a, b.ndim - a.ndim) * b



# Copyright 2023 The autobound Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A library for editing computation graphs.

This library offers more limited functionality than dedicated pattern-matching
libraries like matchpy (https://github.com/HPAC/matchpy).  However, unlike
matchpy and many other existing libraries, it works directly on computation
graphs rather than on trees.  This makes it easier to handle certain use cases,
like editing Jaxprs.
"""

import dataclasses
import itertools
from typing import (Any, Callable, Dict, Hashable, Mapping, Optional, Sequence,
                    Set)

# An intermediate variable in a computation graph (e.g., representing a tensor).
IntermediateVariable = Hashable


@dataclasses.dataclass
class Operation:
  """An operation (a.k.a. equation) in a computation graph."""
  data: Any  # for example, the operation type (e.g., 'MatMul')
  inputs: Sequence[IntermediateVariable]
  outputs: Sequence[IntermediateVariable]

  def subs(self, mapping: Mapping[IntermediateVariable, IntermediateVariable]):
    return Operation(self.data,
                     [mapping.get(u, u) for u in self.inputs],
                     [mapping.get(v, v) for v in self.outputs])


@dataclasses.dataclass
class ComputationGraph:
  """A computation graph (e.g., representing a Jaxpr or a tf.Graph)."""
  inputs: Sequence[IntermediateVariable]
  outputs: Sequence[IntermediateVariable]
  operations: Sequence[Operation]
  data: Any = None  # Any global data associated with the graph.

  def intermediate_variables(self) -> Set[IntermediateVariable]:
    intvars = set(self.inputs)
    intvars.update(self.outputs)
    for op in self.operations:
      intvars.update(op.inputs)
      intvars.update(op.outputs)
    return intvars


def match(
    pattern: Sequence[Operation],
    subject: Sequence[Operation],
    can_bind: Callable[[IntermediateVariable, IntermediateVariable], bool]
) -> Optional[Dict[IntermediateVariable, IntermediateVariable]]:
  """Checks whether a pattern matches a subject, and returns mapping if so.

  Args:
    pattern: a sequence of `Operations`
    subject: a sequence of `Operations`
    can_bind: a callable that, given as arguments a pattern
      `IntermediateVariable` `u` and a subject `IntermediateVariable` `v`,
      determines whether `u` can be mapped to `v`.  (This could return `False`,
      for example, if `u `represents a constant and `v` represents a different
      constant.)

  Returns:
    A dict `m` representing a match, or `None` if no match was found.  If the
    return value is not `None`, it maps from pattern `IntermediateVertex` to
    subject `IntermediateVertex`, and satisfies:
    ```python
    [e.subs(m) for e in pattern] == subject
    ```
  """
  if len(pattern) != len(subject):
    return None

  vertex_map = {}  # from pattern vertex to subject vertex
  for p, s in zip(pattern, subject):
    if (len(p.inputs) != len(s.inputs) or len(p.outputs) != len(s.outputs)):
      return None
    if p.data != s.data:
      return None
    for u, v in itertools.chain(zip(p.inputs, s.inputs),
                                zip(p.outputs, s.outputs)):
      if u in vertex_map:
        if vertex_map[u] != v:
          return None
      else:
        if can_bind(u, v):
          vertex_map[u] = v
        else:
          return None
  return vertex_map


def replace(
    pattern: Sequence[Operation],
    replacement: Sequence[Operation],
    subject: ComputationGraph,
    can_bind: Callable[[IntermediateVariable, IntermediateVariable], bool]
) -> ComputationGraph:
  """Perform a search/replace on a ComputationGraph.

  This method greedily replaces occurences of a given operation sequence
  `pattern` with an operation sequence `replacement`.

  Args:
    pattern: a sequence of `Operations`
    replacement: a sequence of `Operations` with which to replace occurrences of
      the pattern.
    subject: a `ComputationGraph`
    can_bind: a `Callable` with the same meaning as the corresponding argument
      to `match()`.

  Returns:
    a `ComputationGraph` with occurrences of `pattern` replaced by
    `replacement`.
  """
  # Note: this could be made more efficient using the Knuth-Morris-Pratt
  # algorithm.
  k = len(pattern)
  output_operations = []
  i = 0
  while i < len(subject.operations):
    subgraph = subject.operations[i:i+k]
    m = match(pattern, subgraph, can_bind)
    if m is not None:
      output_operations.extend([e.subs(m) for e in replacement])
      i += k
    else:
      output_operations.append(subject.operations[i])
      i += 1

  return ComputationGraph(
      subject.inputs,
      subject.outputs,
      output_operations,
      data=subject.data
  )



# Copyright 2023 The autobound Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The IntervalArithmetic class."""

import functools
from typing import Callable, Sequence, Union

# pylint: disable=g-multiple-import
from autobound.types import (
    Interval, IntervalLike, NDArray, NDArrayLike, NumpyLike)


class IntervalArithmetic:
  """Interval arithmetic on n-dimensional arrays via a numpy-like back end.

  Note: the intervals returned by methods in this class are only correct up to
  floating point roundoff error.
  """

  def __init__(self, np_like: NumpyLike):
    self.np_like = np_like

  def add(self,
          a: Union[NDArrayLike, IntervalLike],
          b: Union[NDArrayLike, IntervalLike]) -> Union[NDArray, Interval]:
    """Returns the sum of two intervals."""
    a_is_interval = isinstance(a, tuple)
    b_is_interval = isinstance(b, tuple)
    if a_is_interval and b_is_interval:
      return (self.np_like.add(a[0], b[0]), self.np_like.add(a[1], b[1]))
    elif a_is_interval:
      return (self.np_like.add(a[0], b), self.np_like.add(a[1], b))
    elif b_is_interval:
      return (self.np_like.add(a, b[0]), self.np_like.add(a, b[1]))
    else:
      return self.np_like.add(a, b)

  def arbitrary_bilinear(
      self,
      a: Union[NDArrayLike, IntervalLike],
      b: Union[NDArrayLike, IntervalLike],
      bilinear: Callable[[NDArrayLike, NDArrayLike], NDArray],
      assume_product: bool = False
  ) -> Union[NDArray, Interval]:
    """Applies a bilinear operation (e.g., matmul, conv2d) to two intervals.

    Args:
      a: an NDArray-like or Interval-like object.
      b: an NDArray-like or Interval-like object.
      bilinear: a callable that takes two NDArray-like arguments, and returns
        the NDArray that results from applying some bilinear operation to them.
      assume_product: if True, we assume that each element of the NDArray
        returned by `bilinear` is a product of some element of `a` and some
        element of `b`, and use a rule that returns a tighter interval under
        this assumption.

    Returns:
      an NDArray or Interval representing the result of applying the bilinear
      operation to `a` and `b`.
    """
    a_is_interval = isinstance(a, tuple)
    b_is_interval = isinstance(b, tuple)
    if not a_is_interval and not b_is_interval:
      return bilinear(a, b)

    if assume_product:
      def yield_endpoint_products():
        a_endpoints = a if a_is_interval else (a,)
        b_endpoints = b if b_is_interval else (b,)
        for a_endpoint in a_endpoints:
          for b_endpoint in b_endpoints:
            yield bilinear(a_endpoint, b_endpoint)  # pytype: disable=wrong-arg-types
      endpoint_products = list(yield_endpoint_products())
      return (
          functools.reduce(self.np_like.minimum, endpoint_products),
          functools.reduce(self.np_like.maximum, endpoint_products)
      )
    else:
      # TODO(mstreeter): there are multiple methods that could be used here,
      # which make different tradeoffs between computation and the tightness of
      # the returned interval.
      # TODO(mstreeter): add reference to proof of correctness for the method
      # used here.
      def positive_and_negative_parts(x):
        return (self.np_like.maximum(0, x), self.np_like.minimum(0, x))

      if not b_is_interval:
        assert a_is_interval
        b_pos, b_neg = positive_and_negative_parts(b)
        min_vals = self.np_like.add(bilinear(a[0], b_pos),
                                    bilinear(a[1], b_neg))
        max_vals = self.np_like.add(bilinear(a[1], b_pos),
                                    bilinear(a[0], b_neg))
        return (min_vals, max_vals)
      elif not a_is_interval:
        a_pos, a_neg = positive_and_negative_parts(a)
        min_vals = self.np_like.add(bilinear(a_pos, b[0]),
                                    bilinear(a_neg, b[1]))
        max_vals = self.np_like.add(bilinear(a_pos, b[1]),
                                    bilinear(a_neg, b[0]))
        return (min_vals, max_vals)
      else:
        assert a_is_interval and b_is_interval
        u, v = a
        w, x = b
        u_pos, u_neg = positive_and_negative_parts(u)
        v_pos, v_neg = positive_and_negative_parts(v)
        w_pos, w_neg = positive_and_negative_parts(w)
        x_pos, x_neg = positive_and_negative_parts(x)
        min_pairs = [(u_pos, w_pos), (v_pos, w_neg),
                     (u_neg, x_pos), (v_neg, x_neg)]
        min_vals = functools.reduce(
            self.np_like.add,
            [bilinear(x, y) for x, y in min_pairs]
        )
        max_pairs = [(v_pos, x_pos), (v_neg, w_pos),
                     (u_pos, x_neg), (u_neg, w_neg)]
        max_vals = functools.reduce(
            self.np_like.add,
            [bilinear(x, y) for x, y in max_pairs]
        )
        return (min_vals, max_vals)

  def as_interval(self, a: IntervalLike) -> Interval:
    return tuple(self.np_like.asarray(c) for c in a)

  def as_interval_or_ndarray(
      self,
      a: Union[NDArrayLike, IntervalLike]) -> Union[NDArray, Interval]:
    if isinstance(a, tuple):
      return self.as_interval(a)
    else:
      return self.np_like.asarray(a)

  def multiply(
      self,
      a: Union[NDArrayLike, IntervalLike],
      b: Union[NDArrayLike, IntervalLike]) -> Union[NDArray, Interval]:
    """Returns the element-wise product of two intervals."""
    return self.arbitrary_bilinear(a, b, self.np_like.multiply, True)

  def ndim(self, a: Union[IntervalLike, NDArrayLike]) -> int:
    return self.np_like.ndim(a[0] if isinstance(a, tuple) else a)

  def negative(self,
               a: Union[NDArrayLike, IntervalLike]) -> Union[NDArray, Interval]:
    """Returns element-wise negative of an interval."""
    if isinstance(a, tuple):
      return (self.np_like.negative(a[1]), self.np_like.negative(a[0]))
    else:
      return self.np_like.negative(a)

  def outer_power(self,
                  a: Union[NDArrayLike, IntervalLike],
                  exponent: int,
                  batch_dims: int = 0) -> Union[NDArray, Interval]:
    """Returns a repeated outer product."""
    if exponent < 0:
      raise ValueError(exponent)
    elif exponent == 0:
      return self.np_like.asarray(1)
    elif self.ndim(a) == 0:
      return self.power(a, exponent)
    else:
      # For the off-diagonal elements of the output, the best we can do is
      # to repeatedly call self.outer_product().  For the diagonal elements,
      # we can use self.power() to get a tighter result.
      a = self.as_interval_or_ndarray(a)
      running_outer_product = a
      for _ in range(exponent - 1):
        running_outer_product = self.outer_product(running_outer_product, a,
                                                   batch_dims)
      if batch_dims != 0:
        # TODO(mstreeter): adapt the code below to handle batch_dims != 0.
        return running_outer_product
      try:
        a_is_interval = isinstance(a, tuple)
        eye = _generalized_diag_ndarray(
            self.np_like.ones_like(a[0] if a_is_interval else a),
            exponent, self.np_like)
      except NotImplementedError:
        return running_outer_product
      diagonal_elements = self.power(a, exponent)
      diag = self._generalized_diag_interval(diagonal_elements, exponent)
      return self.add(
          self.multiply(running_outer_product, 1 - eye),
          diag
      )

  def outer_product(self,
                    a: Union[NDArrayLike, IntervalLike],
                    b: Union[NDArrayLike, IntervalLike],
                    batch_dims: int = 0) -> Union[NDArray, Interval]:
    """Interval variant of _ndarray_outer_product()."""
    if batch_dims > self.ndim(a) or batch_dims > self.ndim(b):
      raise ValueError((self.ndim(a), self.ndim(b), batch_dims))
    product = functools.partial(_ndarray_outer_product,
                                batch_dims=batch_dims, np_like=self.np_like)
    return self.arbitrary_bilinear(a, b, product, True)

  def power(self, a: Union[NDArrayLike, IntervalLike],
            exponent: float) -> Union[NDArray, Interval]:
    """Returns a**exponent (element-wise)."""
    a = self.as_interval_or_ndarray(a)
    a_is_interval = isinstance(a, tuple)
    if a_is_interval:
      if exponent < 0:
        raise NotImplementedError(exponent)
      elif exponent == 0:
        return self.np_like.ones_like(a[0])
      else:
        # For scalars u and v, with u <= v, and even K, the left end point of
        # [u, v]**K is 0 if u <= 0 <= v, and is min{u**K, v**K} otherwise.
        # If K is odd, the left end point is u**K.  The expression for
        # min_vals below handles all cases.
        #
        # The right and point is always max{u**K, v**K}, giving a simpler
        # expression for max_vals.
        contains_zero = self.np_like.logical_and(a[0] < 0, a[1] > 0)
        pow0 = a[0]**exponent
        pow1 = a[1]**exponent
        min_vals = functools.reduce(self.np_like.minimum,
                                    [pow0, pow1, (1-contains_zero)*pow0])
        max_vals = self.np_like.maximum(pow0, pow1)
        return (min_vals, max_vals)
    else:
      return self.np_like.power(a, exponent)

  def shape(self, a: Union[IntervalLike, NDArrayLike]):
    # Note: calling tuple(...) is necessary when self.np_like is
    # tf.experimental.numpy.
    return tuple(self.np_like.shape(a[0] if isinstance(a, tuple) else a))

  def subtract(
      self,
      a: Union[NDArrayLike, IntervalLike],
      b: Union[NDArrayLike, IntervalLike]) -> Union[NDArray, Interval]:
    """Returns the difference between two intervals."""
    a_is_interval = isinstance(a, tuple)
    b_is_interval = isinstance(b, tuple)
    if a_is_interval and b_is_interval:
      return (self.np_like.subtract(a[0], b[1]),
              self.np_like.subtract(a[1], b[0]))
    elif a_is_interval:
      return (self.np_like.subtract(a[0], b), self.np_like.subtract(a[1], b))
    elif b_is_interval:
      return (self.np_like.subtract(a, b[1]), self.np_like.subtract(a, b[0]))
    else:
      return self.np_like.subtract(a, b)

  def tensordot(
      self,
      a: Union[NDArrayLike, IntervalLike],
      b: Union[NDArrayLike, IntervalLike],
      axes) -> Union[NDArray, Interval]:
    """Like np.tensordot(), but for intervals."""
    bilinear = functools.partial(self.np_like.tensordot, axes=axes)
    return self.arbitrary_bilinear(a, b, bilinear, axes == 0)

  def _generalized_diag_interval(
      self,
      a: Union[NDArrayLike, IntervalLike], n: int) -> Union[NDArray, Interval]:
    """Interval variant of _generalized_diag_ndarray."""
    if isinstance(a, tuple):
      if len(a) != 2:
        raise ValueError()
      return (_generalized_diag_ndarray(a[0], n, self.np_like),
              _generalized_diag_ndarray(a[1], n, self.np_like))
    else:
      return _generalized_diag_ndarray(a, n, self.np_like)


def _generalized_diag_ndarray(a: NDArrayLike, n: int,
                              np_like: NumpyLike) -> NDArray:
  """Returns NDArray of shape shape(a)*n, with a on diagonal."""
  if n == 1:
    return np_like.asarray(a)
  elif n == 2:
    a = np_like.asarray(a)
    if a.ndim == 1:
      return np_like.diag(a)
    else:
      raise NotImplementedError(a.ndim)
  else:
    raise NotImplementedError(n)


def _stringify(axes: Sequence[int]) -> str:
  """Helper for creating an einsum() equation string."""
  offset = ord('a')
  return ''.join(chr(i + offset) for i in axes)


def _ndarray_outer_product(a: NDArrayLike,
                           b: NDArrayLike,
                           batch_dims: int,
                           np_like: NumpyLike) -> NDArray:
  """Returns an outer product with batch dimensions.

  Args:
    a: an NDArray-like object.
    b: an NDArray-like object.
    batch_dims: number of batch dimensions.
    np_like: a Numpy-like backend.

  Returns:
    an NDArray c such that, for every tuple I that indexes the first
    batch_dims elements of a (and b), every tuple J that indexes the last
    a.ndim - batch_dims elements of a, and every tuple K that indexes the last
    a.ndim - batch_dims elements of b, we have:

      c[I + J + K] = a[I + J] * b[I + K]
  """
  a = np_like.asarray(a)
  b = np_like.asarray(b)
  if batch_dims == 0:
    return np_like.tensordot(a, b, 0)
  else:
    a_axes = tuple(range(a.ndim))
    b_non_batch_axes = tuple(range(a.ndim, a.ndim + b.ndim - batch_dims))
    b_axes = tuple(range(batch_dims)) + b_non_batch_axes
    output_axes = a_axes + b_non_batch_axes
    eq = (_stringify(a_axes) + ',' + _stringify(b_axes) +
          '->' + _stringify(output_axes))
    return np_like.einsum(eq, a, b)



# Copyright 2023 The autobound Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Library that lists properties of various one-dimensional functions."""

import dataclasses
import functools
import math

from typing import Callable, List, Optional, Sequence, Tuple
# pylint: disable=g-multiple-import
from autobound.types import (Interval, IntervalLike, NDArray, NDArrayLike,
                             NumpyLike)


@dataclasses.dataclass(eq=True, frozen=True)
class FunctionId:
  """An identifier for an one-dimensional function."""
  # Functions are identified by a unique name and a derivative order.  For
  # example, the third derivative of the sigmoid function has name == 'sigmoid'
  # and derivative_order == 3.
  #
  # In some cases, multiple FunctionIDs refer to the same mathematical function.
  # For example, the sigmoid function can be represented by the FunctionId
  # with name == 'sigmoid' and derivative_order == 0, or by the FunctionId
  # with name == 'softplus' and derivative_order == 1.
  name: str
  derivative_order: int = 0
  # x_min and x_max specify the domain of function.  A value of None represents
  # +/- infinity.
  x_min: Optional[float] = None
  x_max: Optional[float] = None

  def derivative_id(self, order: int):
    """Returns `FunctionID` for order `order` derivative of this function."""
    return FunctionId(self.name, self.derivative_order + order,
                      x_min=self.x_min, x_max=self.x_max)


@dataclasses.dataclass(eq=True, frozen=True)
class FunctionData:
  """An object that lists properties of a one-dimensional function."""
  # The lists of local minima and maxima include all minima/maxima over the
  # domain of the function, in ascending order.
  local_minima: Tuple[float, ...]
  local_maxima: Tuple[float, ...]
  monotonically_decreasing: bool = False
  monotonically_increasing: bool = False
  even_symmetric: bool = False  # whether f(x) = f(-x) for all x.

  def monotone_over(
      self,
      region: IntervalLike,
      np_like: NumpyLike) -> Tuple[NDArray, NDArray]:
    """Returns ndarrays showing whether the function is monotone over `region`.

    Args:
      region: an `Interval`
      np_like: a `NumpyLike` back end.

    Returns:
      a pair of boolean `NDArray`s `(decreasing, increasing)`, where the
      elements of `decreasing` (resp `increasing`) indicate whether the
      function is monotonically decreasing (resp increasing) over the interval
      specified by the corresponding elements of `region`.
    """
    x_min = np_like.asarray(region[0])
    x_max = np_like.asarray(region[1])

    sorted_extrema = sorted(self.local_minima + self.local_maxima)
    decreasing_conditions = []
    increasing_conditions = []
    for i, x in enumerate(sorted_extrema):
      is_minimum = x in self.local_minima
      if i == 0:
        if is_minimum:
          decreasing_conditions.append(x_max <= x)
        else:
          increasing_conditions.append(x_max <= x)
      else:
        prev_x = sorted_extrema[i-1]
        contained = np_like.logical_and(x_min >= prev_x, x_max <= x)
        if is_minimum:
          decreasing_conditions.append(contained)
        else:
          increasing_conditions.append(contained)
      if i == len(sorted_extrema) - 1:
        if is_minimum:
          increasing_conditions.append(x_min >= x)
        else:
          decreasing_conditions.append(x_min >= x)

    decreasing = functools.reduce(
        np_like.logical_or, decreasing_conditions,
        np_like.full(x_min.shape, self.monotonically_decreasing))
    increasing = functools.reduce(
        np_like.logical_or, increasing_conditions,
        np_like.full(x_min.shape, self.monotonically_increasing))
    return decreasing, increasing


# FunctionIds for various elementwise functions.
ABS = FunctionId('abs')
EXP = FunctionId('exp')
LOG = FunctionId('log', x_min=0.)
SIGMOID = FunctionId('sigmoid')
SOFTPLUS = FunctionId('softplus')
SWISH = FunctionId('swish')


def get_function(function_id: FunctionId,
                 np_like: NumpyLike) -> Callable[[NDArray], NDArray]:
  """Returns a callable version of a function with a given `FunctionId`."""
  if function_id.name == ABS.name:
    if function_id.derivative_order == 0:
      return np_like.abs
    elif function_id.derivative_order == 1:
      return np_like.sign
    else:
      raise NotImplementedError(function_id.derivative_order)
  if function_id.name == EXP.name:
    return np_like.exp
  elif function_id.name == LOG.name:
    k = function_id.derivative_order
    if k == 0:
      return np_like.log
    else:
      sign = -1 if k % 2 == 0 else 1
      return lambda x: sign * math.factorial(k-1) * np_like.asarray(x)**-k
  elif function_id.name == SIGMOID.name:
    return functools.partial(_sigmoid_derivative,
                             function_id.derivative_order, np_like=np_like)
  elif function_id.name == SOFTPLUS.name:
    return functools.partial(_softplus_derivative,
                             function_id.derivative_order, np_like=np_like)
  elif function_id.name == SWISH.name:
    return functools.partial(_swish_derivative,
                             function_id.derivative_order, np_like=np_like)
  else:
    raise NotImplementedError(function_id)


def get_function_data(function_id: FunctionId) -> FunctionData:
  """Gets `FunctionData` given `FunctionId`."""
  if function_id.name == EXP.name:
    return FunctionData((), (), monotonically_increasing=True)
  elif function_id.name == LOG.name:
    k = function_id.derivative_order
    return FunctionData(
        # The domain of the log function is (0, infinity).  Over this domain,
        # none of the derivatives have any local extrema.
        (),
        (),
        monotonically_decreasing=(k%2 == 1),
        monotonically_increasing=(k%2 == 0)
    )
  elif function_id.name == SIGMOID.name:
    # Sigmoid is 1st derivative of softplus, so kth derivative of sigmoid is
    # (k+1)st derivative of softplus.
    return get_function_data(
        SOFTPLUS.derivative_id(1 + function_id.derivative_order))
  elif function_id in _FUNCTION_DATA:
    return _FUNCTION_DATA[function_id]
  else:
    raise NotImplementedError(function_id)


def get_taylor_polynomial_coefficients(
    function_id: FunctionId,
    degree: int,
    x0: NDArray,
    np_like: NumpyLike) -> List[NDArray]:
  """Returns the Taylor polynomial coefficients for a given function at `x0`.

  Args:
    function_id: a `FunctionId`
    degree: the degree of the Taylor polynomial whose coefficients we return
    x0: the reference point
    np_like: a `NumpyLike` backend.

  Returns:
    a list of `NDArray`s of Taylor polynomial coefficients, of length
    `degree+1`.
  """
  coefficients = []
  for i in range(degree + 1):
    f_deriv = get_function(function_id.derivative_id(i), np_like)
    coefficients.append(f_deriv(x0) / math.factorial(i))
  return coefficients


def maximum_value(f,
                  x_min: NDArray,
                  x_max: NDArray,
                  local_maxima: Sequence[float],
                  np_like: NumpyLike) -> NDArray:
  """Returns maximum value of `f` over `[x_min, x_max]`."""
  if not local_maxima:
    return np_like.maximum(f(x_min), f(x_max))
  sorted_maxima = list(sorted(local_maxima, key=f, reverse=True))
  x = sorted_maxima[0]
  return np_like.where(
      np_like.logical_and(x_min <= x, x <= x_max),
      f(x),
      maximum_value(f, x_min, x_max, sorted_maxima[1:], np_like))


def minimum_value(f,
                  x_min: NDArray,
                  x_max: NDArray,
                  local_minima: Sequence[float],
                  np_like: NumpyLike) -> NDArray:
  """Returns minimum value of `f` over `[x_min, x_max]`."""
  if not local_minima:
    return np_like.minimum(f(x_min), f(x_max))
  sorted_minima = list(sorted(local_minima, key=f))
  x = sorted_minima[0]
  return np_like.where(
      np_like.logical_and(x_min <= x, x <= x_max),
      f(x),
      minimum_value(f, x_min, x_max, sorted_minima[1:], np_like))


def _get_range(f,
               x_min: NDArray,
               x_max: NDArray,
               local_minima: Sequence[float],
               local_maxima: Sequence[float],
               np_like: NumpyLike) -> Tuple[NDArray, NDArray]:
  minval = minimum_value(f, x_min, x_max, local_minima, np_like)
  maxval = maximum_value(f, x_min, x_max, local_maxima, np_like)
  return (minval, maxval)


def get_range(function_id: FunctionId,
              trust_region: Interval,
              np_like: NumpyLike) -> Interval:
  """Returns exact range of specified function over `trust_region`."""
  f = get_function(function_id, np_like)
  function_data = get_function_data(function_id)
  return _get_range(
      f,
      trust_region[0],
      trust_region[1],
      function_data.local_minima,
      function_data.local_maxima,
      np_like
  )


def _sigmoid(x: NDArrayLike, np_like: NumpyLike) -> NDArray:
  return np_like.where(
      x >= 0,
      1 / (1 + np_like.exp(-x)),
      np_like.exp(x) / (1 + np_like.exp(x))
  )


def _sigmoid_derivative(order: int, x: NDArrayLike,
                        np_like: NumpyLike) -> NDArray:
  """Returns the (elementwise) derivative of a specified order."""
  # Note: we could make this work for arbitrary order using autodiff, but we
  # don't because this module is backend-agnostic, and we don't have a way to
  # do autodiff in a backend-agnostic way.
  s = _sigmoid(x, np_like)
  sm = _sigmoid(-x, np_like)
  if order == 0:
    return s
  elif order == 1:
    return s*sm
  elif order == 2:
    return s*sm*(1-2*s)
  elif order == 3:
    return s*sm*((1-2*s)**2 - 2*s*sm)
  elif order == 4:
    return (s*sm*(1-2*s)*((1-2*s)**2 - 2*s*sm) +
            s*sm*(-4*(1-2*s)*s*sm - 2*s*sm*(1-2*s)))
  else:
    raise NotImplementedError(order)


def _softplus(x: NDArrayLike, np_like: NumpyLike) -> NDArray:
  # Avoid overflow for large positive x using:
  # log(1+exp(x)) == log(1+exp(-|x|)) + max(x, 0).
  return np_like.log1p(np_like.exp(-np_like.abs(x))) + np_like.maximum(x, 0)


def _softplus_derivative(order: int, x: NDArrayLike,
                         np_like: NumpyLike) -> NDArray:
  if order == 0:
    return _softplus(x, np_like)
  else:
    return _sigmoid_derivative(order - 1, x, np_like)


def _swish(x: NDArrayLike, np_like: NumpyLike) -> NDArray:
  return x*_sigmoid(x, np_like)


def _swish_derivative(order: int, x: NDArrayLike,
                      np_like: NumpyLike) -> NDArray:
  if order == 0:
    return _swish(x, np_like)
  else:
    # swish(x) = x*sigmoid(x)
    # swish'(x) = sigmoid(x) + x*sigmoid'(x).
    # Inductively,
    # swish^{(k)}(x) = k*sigmoid^({k-1})(x) + x*sigmoid^{(k)}(x).
    return (order*_sigmoid_derivative(order - 1, x, np_like) +
            x*_sigmoid_derivative(order, x, np_like))


# Dict from FunctionId to FunctionData.
_FUNCTION_DATA = {
    # ABS
    ABS: FunctionData((0.,), (), even_symmetric=True),
    ABS.derivative_id(1): FunctionData((), (), monotonically_increasing=True),
    # SOFTPLUS
    SOFTPLUS: FunctionData((), (), monotonically_increasing=True),
    SOFTPLUS.derivative_id(1):
        FunctionData((), (), monotonically_increasing=True),
    SOFTPLUS.derivative_id(2):
        FunctionData((), (0.,), even_symmetric=True),
    SOFTPLUS.derivative_id(3):
        FunctionData((1.3169578969249405,), (-1.3169578969249423,)),
    SOFTPLUS.derivative_id(4):
        FunctionData((0.,), (-2.292431669561122, 2.2924316695611195)),
    SOFTPLUS.derivative_id(5):
        FunctionData((-0.8426329481295408, 3.1443184061547065),
                     (-3.144318406154709, 0.8426329481295388)),
    # SWISH
    SWISH: FunctionData((-1.278464542761141,), ()),
    SWISH.derivative_id(1):
        FunctionData((-2.399357280515326,), (2.399357280515324,)),
    SWISH.derivative_id(2):
        FunctionData((-3.4358409935350243, 3.4358409935350225), (0.,)),
    SWISH.derivative_id(3):
        FunctionData((-4.429235100557346, 1.0319582417807385),
                     (-1.0319582417807402, 4.429235100557342)),
    SWISH.derivative_id(4):
        FunctionData((0.,), (-1.8197756117249821, 1.8197756117249804)),
    SWISH.derivative_id(5):
        FunctionData((-0.7177419231466055, 2.5062894864026024),
                     (-2.506289486402605, 0.7177419231466035)),
}



# Copyright 2023 The autobound Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A library of Taylor enclosures for various primitive functions.

For now this only supports elementwise functions, but in the future it will
support other multivariate functions.
"""

import functools
import math
from typing import Callable, Sequence, Tuple

from autobound import elementwise_functions
# pylint: disable=g-multiple-import
from autobound.types import (
    Interval, NumpyLike, ElementwiseTaylorEnclosure, NDArray, NDArrayLike)


def get_elementwise_taylor_enclosure(
    function_id: elementwise_functions.FunctionId,
    x0: NDArray,
    trust_region: Interval,
    degree: int,
    np_like: NumpyLike) -> ElementwiseTaylorEnclosure:
  """Returns ElementwiseTaylorEnclosure for function with given ID.

  Args:
    function_id: an `elementwise_functions.FunctionId`
    x0: reference point
    trust_region: trust region over which enclosure is valid
    degree: the degree of the returned `ElementwiseTaylorEnclosure`
    np_like: a `NumpyLike` backend

  Returns:
    an `ElementwiseTaylorEnclosure` for the elementwise function specified by
    `function_id`.
  """
  f = elementwise_functions.get_function(function_id, np_like)
  deriv_id = function_id.derivative_id(degree)
  deriv_data = elementwise_functions.get_function_data(deriv_id)
  taylor_coefficients = functools.partial(
      elementwise_functions.get_taylor_polynomial_coefficients,
      function_id, x0=x0, np_like=np_like)
  if (deriv_data.monotonically_increasing or
      deriv_data.monotonically_decreasing):
    return sharp_enclosure_monotonic_derivative(
        x0, degree, trust_region, f, taylor_coefficients(degree),
        deriv_data.monotonically_increasing, np_like)
  elif degree == 2 and deriv_data.even_symmetric:
    return sharp_quadratic_enclosure_even_symmetric_hessian(
        x0, trust_region, f, taylor_coefficients(degree), np_like)
  else:
    # For indices where the derivative is monotonically decreasing or
    # monotonically increasing over the trust region, we return the sharp
    # enclosure.  For other indices, we fall back to using the enclosure
    # based on the range of the derivative.
    coeffs = taylor_coefficients(degree)
    enclosure_if_decreasing, enclosure_if_increasing = [
        sharp_enclosure_monotonic_derivative(
            x0, degree, trust_region, f, coeffs, increasing, np_like)
        for increasing in [False, True]
    ]
    decreasing, increasing = deriv_data.monotone_over(trust_region, np_like)
    deriv_range = elementwise_functions.get_range(deriv_id, trust_region,
                                                  np_like)
    fallback = bounded_derivative_enclosure(degree, coeffs[:-1], deriv_range)
    def endpoint(i: int):
      return np_like.where(
          decreasing,
          enclosure_if_decreasing[-1][i],
          np_like.where(
              increasing,
              enclosure_if_increasing[-1][i],
              fallback[-1][i]
          )
      )
    final_interval = (endpoint(0), endpoint(1))
    return ElementwiseTaylorEnclosure(
        tuple(coeffs[:degree]) + (final_interval,))


abs_enclosure = functools.partial(get_elementwise_taylor_enclosure,
                                  elementwise_functions.ABS)
exp_enclosure = functools.partial(get_elementwise_taylor_enclosure,
                                  elementwise_functions.EXP)
log_enclosure = functools.partial(get_elementwise_taylor_enclosure,
                                  elementwise_functions.LOG)
sigmoid_enclosure = functools.partial(get_elementwise_taylor_enclosure,
                                      elementwise_functions.SIGMOID)
softplus_enclosure = functools.partial(get_elementwise_taylor_enclosure,
                                       elementwise_functions.SOFTPLUS)
swish_enclosure = functools.partial(get_elementwise_taylor_enclosure,
                                    elementwise_functions.SWISH)


# TODO(mstreeter): we could implement pow_enclosure in terms of
# get_elementwise_taylor_enclosure if we allowed FunctionIds to have parameters
# (in this case, the exponent).
def pow_enclosure(exponent: float,
                  x0: NDArray,
                  trust_region: Interval,
                  degree: int,
                  np_like: NumpyLike) -> ElementwiseTaylorEnclosure:
  """Returns an ElementwiseTaylorEnclosure for x**exponent in terms of x-x0."""
  # The kth derivative of x**p is p * (p-1) * ... * (p-k) * x0**(p-k)
  taylor_coefficients_at_x0 = []
  c = 1.
  i_factorial = 1.
  for i in range(degree + 1):
    if i > 0:
      i_factorial *= i
    # Note: the next line can sometimes generate bogus RuntimeWarnings when
    # using Numpy.  This seems to be a bug in Numpy, as even doing
    # np.array(2.)**-1 generates the same RuntimeWarning.
    taylor_coefficients_at_x0.append(c * x0**(exponent - i) / i_factorial)
    if i < degree:
      c *= exponent - i

  # Compute sharp enclosures for two cases: x-x0 > 0 (enc_pos below), and
  # x-x0 < 0 (enc_neg below).
  #
  # The kth derivative of x**p at x is c*x**(p-k), where c
  # is negative if k is odd, and positive if k is even.
  # For x > 0, c*x**(p-k) is decreasing if c is positive, and
  # decreasing otherwise.
  # For x < 0 and even p-k, the situation is the same.
  # For x < 0 and odd p-k, the situation is reversed: c*x**(p-k) is increasing
  # if c is positive and decreasing otherwise.
  # pylint: disable=g-complex-comprehension
  enc_decreasing, enc_increasing = [
      sharp_enclosure_monotonic_derivative(
          x0, degree, trust_region, lambda x: x**exponent,
          taylor_coefficients_at_x0, increasing, np_like
      )
      for increasing in [False, True]
  ]
  enc_pos = enc_decreasing if c > 0 else enc_increasing
  if exponent - degree % 2 == 0:
    enc_neg = enc_pos
  else:
    enc_neg = enc_increasing if c > 0 else enc_decreasing

  def interval_endpoint(i):
    """Returns left (i == 0) or right (i == 1) endpoint of interval."""
    a, b = trust_region
    endpoint_if_positive = enc_pos[-1][i]
    if int(exponent) != exponent:
      # If exponent is not an integer, then z**exponent is undefined for z < 0.
      # We return the interval (-inf, inf) in this case.
      endpoint_if_negative = -np_like.inf if i == 0 else np_like.inf
      endpoint_if_possibly_zero = endpoint_if_negative
    elif exponent < 0:
      endpoint_if_negative = enc_neg[-1][i]
      endpoint_if_possibly_zero = -np_like.inf if i == 0 else np_like.inf
    else:
      endpoint_if_negative = enc_neg[-1][i]
      endpoint_if_possibly_zero = functools.reduce(
          np_like.minimum if i == 0 else np_like.maximum,
          [endpoint_if_positive, endpoint_if_negative]
      )
    return np_like.where(
        a >= 0,
        endpoint_if_positive,
        np_like.where(
            b <= 0,
            endpoint_if_negative,
            endpoint_if_possibly_zero
        )
    )

  interval_coefficient = tuple(interval_endpoint(i) for i in [0, 1])
  return ElementwiseTaylorEnclosure(
      enc_decreasing[:-1] + (interval_coefficient,))


def bounded_derivative_enclosure(
    degree: int,
    taylor_coefficients_at_x0: Sequence[NDArray],
    derivative_bound: Tuple[NDArray, NDArray]
) -> ElementwiseTaylorEnclosure:
  if len(taylor_coefficients_at_x0) != degree:
    raise ValueError()
  degree_factorial = math.factorial(degree)
  final_interval = (derivative_bound[0] / degree_factorial,
                    derivative_bound[1] / degree_factorial)
  return ElementwiseTaylorEnclosure(
      tuple(taylor_coefficients_at_x0[:degree]) + (final_interval,)
  )


def sharp_enclosure_monotonic_derivative(
    x0: NDArray,
    degree: int,
    trust_region: Interval,
    sigma: Callable[[NDArray], NDArray],
    taylor_coefficients_at_x0: Sequence[NDArray],
    increasing: bool,
    np_like: NumpyLike
) -> ElementwiseTaylorEnclosure:
  """Returns sharp degree-k enclosure assuming monotone k-th derivative.

  Args:
    x0: the center point for the Taylor enclosure
    degree: the degree of the enclosure to return
    trust_region: the trust region over which to compute an enclosure
    sigma: the function for which to compute a sharp polynomial enclosure
    taylor_coefficients_at_x0: the first (degree+1) coefficients of the
      Taylor series expansion of sigma at x0.
    increasing: whether the (degree)th derivative of sigma is increasing
      or decreasing
    np_like: a NumpyLike backend

  Returns:
    a sharp ElementwiseTaylorEnclosure for sigma
  """
  if degree < 0:
    raise ValueError(degree)
  ratio = functools.partial(taylor_remainder_ratio,
                            x0, degree, sigma,
                            taylor_coefficients_at_x0,
                            np_like=np_like)
  a, b = trust_region
  if increasing:
    final_interval = (ratio(a), ratio(b))
  else:
    final_interval = (ratio(b), ratio(a))
  return ElementwiseTaylorEnclosure(
      tuple(taylor_coefficients_at_x0[:degree]) + (final_interval,)
  )


def sharp_quadratic_enclosure_even_symmetric_hessian(
    x0: NDArray,
    trust_region: Interval,
    sigma: Callable[[NDArray], NDArray],
    taylor_coefficients_at_x0: Sequence[NDArray],
    np_like: NumpyLike
) -> ElementwiseTaylorEnclosure:
  """Returns sharp quadratic enclosure for function with even-symmetric Hessian.

  It's assumed that the Hessian is decreasing at z >= 0.

  Args:
    x0: the center point for the Taylor enclosure
    trust_region: the trust region over which to compute an enclosure
    sigma: an elementwise function for which to compute a Taylor enclosure
    taylor_coefficients_at_x0: the first two coefficients of the
      Taylor series expansion of sigma at x0.
    np_like: a Numpy-like back end.
  """
  ratio = functools.partial(taylor_remainder_ratio,
                            x0, 2, sigma,
                            taylor_coefficients_at_x0,
                            np_like=np_like)

  a, b = trust_region
  max_ratio = ratio(np_like.clip(-x0, a, b))
  min_ratio = np_like.minimum(ratio(a), ratio(b))
  final_interval = (min_ratio, max_ratio)
  return ElementwiseTaylorEnclosure(
      tuple(taylor_coefficients_at_x0[:2]) + (final_interval,)
  )


def taylor_remainder_ratio(
    x0: NDArray,
    degree: int,
    sigma: Callable[[NDArray], NDArray],
    taylor_coefficients_at_x0: Sequence[NDArray],
    x: NDArray,
    np_like: NumpyLike):
  """Returns R_{degree - 1}(x; sigma, x0) / (x - x0)**degree."""
  if len(taylor_coefficients_at_x0) != degree + 1:
    raise ValueError(degree, taylor_coefficients_at_x0)
  # Let r_k denote the degree k Taylor series remainder.
  #
  # Letting k = degree, we want to return r_{k-1} / (x-x0)**k, but in a way that
  # is numerically stable when x-x0 is small (and that is well-defined when
  # x=x0).
  #
  # We do so using:
  #     r_{k-1} / (x-x0)**k = (r_k + c_k*(x-x0)**k) / (x-x0)**k
  #                         = c_k + r_k / (x-x0)**k.
  r_k = sigma(x) - sum(
      c * (x - x0)**i for i, c in enumerate(taylor_coefficients_at_x0))
  denom = (x-x0)**degree
  return (
      taylor_coefficients_at_x0[degree] +
      # Return r_k * 1 / denom, capping the magnitude of 1 / denom at 1e12.
      # TODO(mstreeter): this results in an enclosure that's  not strictly valid
      # when denom is very small.
      r_k * np_like.sign(denom) / (np_like.maximum(1e-12, np_like.abs(denom)))
  )



# Copyright 2023 The autobound Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Code for computing Taylor enclosures in JAX."""

import dataclasses
import functools
from typing import Callable, Dict, List, Optional, Set, Tuple, Union

from autobound import enclosure_arithmetic
from autobound import interval_arithmetic
from autobound import polynomials
from autobound import primitive_enclosures
from autobound import types
from autobound.jax import jaxpr_editor
import jax
import jax.numpy as jnp


@dataclasses.dataclass
class TaylorBounds:
  """Upper and lower bounds on a function, valid over a trust region."""
  f: Callable[[jnp.array], jnp.array]
  x0: jnp.ndarray
  # Interval containing values of x (not x-x0) for which the bound on f(x)
  # holds.
  x_trust_region: Tuple[jnp.array, jnp.array]
  coefficients: types.TaylorEnclosure

  def __call__(self,
               x: types.NDArrayLike) -> Union[types.NDArray, types.Interval]:
    x = jnp.asarray(x)
    return polynomials.eval_taylor_enclosure(self.coefficients, x-self.x0, jnp)

  def final_interval(self) -> Tuple[jnp.array, jnp.array]:
    """Returns final coefficient (as a trivial interval, if it is scalar)."""
    c = self.coefficients[-1]
    return c if isinstance(c, tuple) else (c, c)

  def lower(self, x):
    bound = self(x)
    return bound[0] if isinstance(bound, tuple) else bound

  def upper(self, x):
    bound = self(x)
    return bound[1] if isinstance(bound, tuple) else bound


def taylor_bounds(
    f: Callable[[jnp.array], jnp.array],
    max_degree: int,
    propagate_trust_regions: bool = False,
) -> Callable[[jnp.array, Tuple[jnp.array, jnp.array]], TaylorBounds]:
  """Returns version of f that returns a TaylorBounds object.

  Args:
    f: a function that takes a jnp.array as input, and returns a jnp.array
    max_degree: the maximum degree TaylorEnclosure for the returned function
      to return
    propagate_trust_regions: if True, trust regions are propagated
      through the Jaxpr, rather than being computed from higher-degree
      enclosures.  This results in tighter bounds at the cost of additional
      memory.

  Returns:
    a function that takes as input a jnp.array x0, and a trust region
    (min_vals, max_vals), and return a TaylorBounds object `bound` such that
    `bound.coefficients` is a TaylorEnclosure g of degree at most max_degree,
    such that:

        f(x) in g(x-x0) for all x with min_vals <= x <= max_vals
  """
  if max_degree < 0:
    raise ValueError(max_degree)
  if max_degree == 0:
    propagate_trust_regions = False  # avoid redundant computation

  jaxpr_factory = jax.make_jaxpr(f)
  def bound_fun(x0: jnp.array,
                x_trust_region: types.Interval) -> TaylorBounds:
    trust_region = interval_arithmetic.IntervalArithmetic(jnp).subtract(
        x_trust_region, x0)

    arithmetic = enclosure_arithmetic.TaylorEnclosureArithmetic(
        max_degree, trust_region, jnp)
    primitive_to_enclosure_fun = _pushforward_funs(arithmetic)

    degree_0_arithmetic = (
        enclosure_arithmetic.TaylorEnclosureArithmetic(0, trust_region, jnp))
    primitive_to_enclosure_fun0 = _pushforward_funs(degree_0_arithmetic)

    closed_jaxpr = jaxpr_factory(x0)
    jaxpr = _rewrite_jaxpr(closed_jaxpr.jaxpr)

    x0 = jnp.asarray(x0)
    if x0.ndim == 0:
      identity = jnp.asarray(1.)
    elif x0.ndim == 1:
      identity = jnp.eye(x0.shape[0])
    else:
      raise NotImplementedError(x0.ndim)
    x0_enclosure = types.TaylorEnclosure(
        (x0, identity) if max_degree > 0 else (x_trust_region,))
    assert len(closed_jaxpr.consts) == len(jaxpr.constvars)
    var_to_intermediate = {
        var: _constant_intermediate_enclosure(val)
        for var, val in zip(jaxpr.constvars, closed_jaxpr.consts)
    }
    assert len(jaxpr.invars) == 1
    var_to_intermediate[jaxpr.invars[0]] = _IntermediateEnclosure(
        enclosure=x0_enclosure,
        trust_region=x_trust_region if propagate_trust_regions else None
    )

    def get_intermediate(
        invar: Union[jax.core.Var, jax.core.Literal]) -> _IntermediateEnclosure:
      if isinstance(invar, jax.core.Var):
        return var_to_intermediate[invar]
      else:
        assert isinstance(invar, jax.core.Literal)
        return _constant_intermediate_enclosure(invar.val)

    for eqn in jaxpr.eqns:
      invar_intermediates = [get_intermediate(invar) for invar in eqn.invars]
      has_non_constant_invars = any(not intermediate.is_constant()
                                    for intermediate in invar_intermediates)
      if has_non_constant_invars:
        fun = primitive_to_enclosure_fun.get(eqn.primitive)
        if fun is None:
          raise NotImplementedError(eqn.primitive)
        outvar_enclosures = fun(*invar_intermediates, **eqn.params)
        if len(eqn.outvars) == 1:
          outvar_enclosures = (outvar_enclosures,)
        if propagate_trust_regions:
          fun0 = primitive_to_enclosure_fun0.get(eqn.primitive)
          assert fun0 is not None
          assert all(i.trust_region is not None for i in invar_intermediates)
          invar_degree_0_intermediates = [
              _IntermediateEnclosure(
                  enclosure=types.TaylorEnclosure((intermediate.trust_region,)))
              for intermediate in invar_intermediates
          ]
          outvar_degree_0_enclosures_a = fun0(*invar_degree_0_intermediates,
                                              **eqn.params)
          if len(eqn.outvars) == 1:
            outvar_degree_0_enclosures_a = [outvar_degree_0_enclosures_a]
          assert len(outvar_degree_0_enclosures_a) == len(outvar_enclosures)
          outvar_degree_0_enclosures_b = [
              enclosure_arithmetic.enclose_enclosure(enclosure, trust_region,
                                                     0, jnp)
              for enclosure in outvar_enclosures
          ]
          outvar_trust_regions = [
              _intersect_intervals(a[0], b[0])
              for a, b in zip(outvar_degree_0_enclosures_a,
                              outvar_degree_0_enclosures_b)
          ]
          for i, (a, b) in enumerate(outvar_trust_regions):
            # It should always be the case that the actual value of the ith
            # output of a function (y0 below) is inside the associated trust
            # region.  But this invariant may not hold due to floating
            # point roundoff error, so we enforce it here.
            #
            # TODO(mstreeter): add a test case that fails if we remove this.
            y0 = outvar_enclosures[i][0]
            outvar_trust_regions[i] = (jnp.minimum(y0, a), jnp.maximum(y0, b))  # pytype: disable=wrong-arg-types
        else:
          outvar_trust_regions = (None,) * len(outvar_enclosures)
        assert all(isinstance(v, tuple) for v in outvar_enclosures), (
            eqn.primitive, fun, outvar_enclosures)
        outvar_intermediates = tuple(
            _IntermediateEnclosure(enclosure=e, trust_region=r)
            for r, e in zip(outvar_trust_regions, outvar_enclosures)
        )
      else:
        invar_values = tuple(intermediate.constant_value()
                             for intermediate in invar_intermediates)
        vals = eqn.primitive.bind(*invar_values, **eqn.params)
        if len(eqn.outvars) == 1:
          vals = (vals,)
        outvar_intermediates = [_constant_intermediate_enclosure(v)
                                for v in vals]

      assert len(outvar_intermediates) == len(eqn.outvars), (
          eqn.primitive, len(outvar_intermediates), len(eqn.outvars))
      for var, intermediate in zip(eqn.outvars, outvar_intermediates):
        assert var not in var_to_intermediate
        assert isinstance(intermediate.enclosure, tuple), (
            eqn.primitive, intermediate)
        _validate_taylor_enclosure(intermediate.enclosure, x0.shape)
        var_to_intermediate[var] = intermediate

    assert len(jaxpr.outvars) == 1
    output_intermediate = get_intermediate(jaxpr.outvars[0])
    return TaylorBounds(f=f, x0=x0, x_trust_region=x_trust_region,
                        coefficients=output_intermediate.enclosure)

  return bound_fun


# Type for functions that generate enclosures for primitive elementwise
# functions.  The callable takes arguments x0, trust_region, degree, and
# np_like, and returns an ElementwiseTaylorEnclosure (see examples in
# primitive_enclosures.py).
ElementwiseEnclosureGeneratingFunction = Callable[
    [types.NDArray, types.Interval, int, types.NumpyLike],
    types.ElementwiseTaylorEnclosure
]


# TODO(mstreeter): add a mechanism for supporting a new elementwise function
# given its FunctionData.
def register_elementwise_primitive(
    p: jax.core.Primitive,
    get_enclosure: ElementwiseEnclosureGeneratingFunction):
  """Register an enclosure-generating function for a user-defined primitive.

  Args:
    p: a jax.core.Primitive
    get_enclosure: an ElementwiseEnclosureGeneratingFunction for p.
  """
  _ELEMENTWISE_PRIMITIVE_ENCLOSURES[p] = get_enclosure


#
# Private variables/functions.
#


_PRIMITIVE_NAMES = set()  # type: Set[str]
# Rewrite rules are callables that return (pattern Jaxpr, replacement Jaxpr)
# pairs.  We make them callables because the Jaxpr returned by jax.make_jaxpr
# depends on how Jax is configured (in particular whether float64 is enabled),
# and we need to use the Jaxprs that match whatever configuration is being
# used when the rule is applied.
_JAXPR_REWRITE_RULES = [
]  # type: List[Callable[[], Tuple[jax.core.Jaxpr, jax.core.Jaxpr]]]


def _register_elementwise_function(
    f: Callable[[jnp.array], jnp.array],
    get_enclosure: ElementwiseEnclosureGeneratingFunction
):
  """Register enclosure-generating function for elementwise Jax function."""
  name = f'__autobound_{f.__name__}__'
  if name in _PRIMITIVE_NAMES:
    raise ValueError(f)
  _PRIMITIVE_NAMES.add(name)
  p = jax.core.Primitive(name)
  p.def_abstract_eval(
      lambda x: jax.abstract_arrays.ShapedArray(x.shape, x.dtype))
  rule = lambda: (jax.make_jaxpr(f)(0.).jaxpr, jax.make_jaxpr(p.bind)(0.).jaxpr)
  _JAXPR_REWRITE_RULES.append(rule)
  register_elementwise_primitive(p, get_enclosure)


# Dict from Jax Primitive to ElementwiseEnclosureGeneratingFunction.
# TODO(mstreeter): support more elementwise functions.
_ELEMENTWISE_PRIMITIVE_ENCLOSURES = {
    jax.lax.abs_p: primitive_enclosures.abs_enclosure,
    jax.lax.exp_p: primitive_enclosures.exp_enclosure,
    jax.lax.log_p: primitive_enclosures.log_enclosure,
}
_register_elementwise_function(jax.nn.sigmoid,
                               primitive_enclosures.sigmoid_enclosure)
_register_elementwise_function(jax.nn.softplus,
                               primitive_enclosures.softplus_enclosure)
_register_elementwise_function(jax.nn.swish,
                               primitive_enclosures.swish_enclosure)


# Set of primitives that can be applied separately to each coefficient of a
# TaylorEnclosure.
_PASS_THRU_PRIMITIVES = frozenset([
    jax.lax.convert_element_type_p,
    jax.lax.reshape_p,
    jax.lax.reduce_sum_p,
    jax.lax.reduce_window_sum_p,
    jax.lax.squeeze_p,
    jax.lax.transpose_p,
    # TODO(mstreeter): add more of these
])


def _rewrite_jaxpr(jaxpr: jax.core.Jaxpr) -> jax.core.Jaxpr:
  """Rewrite a Jaxpr to make is suitable for use by taylor_bounds()."""
  for rule_generator in _JAXPR_REWRITE_RULES:
    pattern, replacement = rule_generator()
    jaxpr = jaxpr_editor.replace(pattern, replacement, jaxpr)
  return jaxpr


@dataclasses.dataclass
class _IntermediateEnclosure:
  """An enclosure for some intermediate variable in a Jaxpr."""
  enclosure: types.TaylorEnclosure
  trust_region: Optional[types.Interval] = None

  def is_constant(self) -> bool:
    """Returns whether self.enclosure represents a constant value."""
    return len(self.enclosure) == 1 and not isinstance(self.enclosure[0], tuple)

  def constant_value(self) -> types.NDArray:
    if not self.is_constant():
      raise ValueError()
    else:
      return self.enclosure[0]  # pytype: disable=bad-return-type


def _broadcast_in_dim_pushforward_fun(intermediate, shape,
                                      broadcast_dimensions):
  """Enclosure-generating function for jax.lax.broadcast_in_dim."""
  enclosure = intermediate.enclosure
  x0 = enclosure[0]
  if isinstance(x0, tuple):
    x0 = enclosure[0][0]
  x_shape = (() if len(enclosure) == 1 else enclosure[1].shape[x0.ndim:])
  def broadcast_ndarray(a, i):
    return jax.lax.broadcast_in_dim(a, shape + i*x_shape, broadcast_dimensions)
  def broadcast_ndarray_or_interval(a, i):
    if isinstance(a, tuple):
      return tuple(broadcast_ndarray(x, i) for x in a)
    else:
      return broadcast_ndarray(a, i)
  return tuple(
      broadcast_ndarray_or_interval(coeff, i)
      for i, coeff in enumerate(enclosure)
  )


def _constant_intermediate_enclosure(val: types.NDArray):
  return _IntermediateEnclosure(enclosure=types.TaylorEnclosure((val,)),
                                trust_region=(val, val))


def _conv_general_dilated_pushforward_fun(arithmetic):
  """Returns function that implements conv_general_dilated on enclosures."""
  def fun(lhs_intermediate: _IntermediateEnclosure,
          rhs_intermediate: _IntermediateEnclosure,
          **params):
    def pairwise_batched_bilinear(a: jnp.array, b: jnp.array,
                                  p: int, q: int) -> jnp.array:
      def move_last_n_dims_to_front(x: jnp.array, n: int):
        if n == 0:
          return x
        perm = tuple(range(x.ndim - n, x.ndim)) + tuple(range(x.ndim - n))
        transposed = jnp.transpose(x, axes=perm)
        return jnp.reshape(transposed, (-1,) + x.shape[:x.ndim-n])

      a_reshaped = move_last_n_dims_to_front(a, p)
      b_reshaped = move_last_n_dims_to_front(b, q)

      c = jax.lax.conv_general_dilated_p.bind(a_reshaped, b_reshaped, **params)
      if p == 0 and q == 0:
        return c
      elif p == 0 or q == 0:
        raise NotImplementedError((p, q))
      c_perm = tuple(range(2, c.ndim)) + (0, 1)
      c_transposed = jnp.transpose(c, axes=c_perm)
      return jnp.reshape(c_transposed,
                         c.shape[2:] + a.shape[a.ndim-p:] + b.shape[b.ndim-q:])

    return arithmetic.arbitrary_bilinear(
        lhs_intermediate.enclosure,
        rhs_intermediate.enclosure,
        pairwise_batched_bilinear)
  return fun


def _dot_general_pushforward_fun(arithmetic):
  """Returns function that implements dot_general on enclosures."""
  def fun(lhs_intermediate: _IntermediateEnclosure,
          rhs_intermediate: _IntermediateEnclosure,
          **params):
    a_contracting_dims = set(a  # pylint: disable=g-complex-comprehension
                             for t in params['dimension_numbers']
                             for a in t[0])
    def pairwise_batched_bilinear(a: jnp.array, b: jnp.array,
                                  p: int, q: int) -> jnp.array:
      transposed_output = jax.lax.dot_general_p.bind(a, b, **params)
      p_start = a.ndim - p - len(a_contracting_dims)
      assert p_start >= 0
      n = transposed_output.ndim
      # Shift axes p_start through p_start+p to the right, so that they start
      # at position n-q.
      assert p_start + p <= n-q, (p_start, p, q, n)
      perm = (
          tuple(range(p_start)) +
          tuple(range(p_start+p, n-q)) +
          tuple(range(p_start, p_start + p)) +
          tuple(range(n-q, n))
      )
      assert len(set(perm)) == n, (p_start, p, q, n, perm)
      return jnp.transpose(transposed_output, axes=perm)
    return arithmetic.arbitrary_bilinear(
        lhs_intermediate.enclosure,
        rhs_intermediate.enclosure,
        pairwise_batched_bilinear
    )
  return fun


def _elementwise_pushforward_fun(arithmetic, get_enclosure):
  f = arithmetic.get_elementwise_fun(get_enclosure)
  def g(intermediate):
    return f(intermediate.enclosure, intermediate.trust_region)
  return g


def _intersect_intervals(
    a: types.Interval, b: types.Interval) -> types.Interval:
  if not len(a) == len(b) == 2:
    raise ValueError()
  return (jnp.maximum(a[0], b[0]), jnp.minimum(a[1], b[1]))  # pytype: disable=wrong-arg-types


def _pass_thru_pushforward_fun(primitive):
  def fun(intermediate, **params):
    return enclosure_arithmetic.map_over_enclosure(
        intermediate.enclosure,
        functools.partial(primitive.bind, **params)
    )
  return fun


# A pushforward function for an underlying primitive with K inputs and N
# outputs takes K _IntermediateEnclosure as arguments, plus kwargs for any
# parameters the primitive has, and returns a tuple of N TaylorEnclosures (or
# in the special case N=1, a single TaylorEnclosure rather than a tuple).
PushforwardFunction = Callable[
    ...,
    Union[types.TaylorEnclosure, Tuple[types.TaylorEnclosure, ...]]
]


def _pushforward_funs(
    arithmetic: enclosure_arithmetic.TaylorEnclosureArithmetic
) -> Dict[jax.core.Primitive, PushforwardFunction]:
  """Returns dict from primitive to function that inputs/outputs enclosures."""
  def pushforward_integer_pow(intermediate, y: int):
    return arithmetic.power(intermediate.enclosure, y)

  def pushforward_pow(intermediate_0, intermediate_1):
    if not intermediate_1.is_constant():
      raise NotImplementedError()
    exponent = float(intermediate_1.constant_value())
    return arithmetic.power(intermediate_0.enclosure, exponent)

  def wrap(f):
    def g(*args):
      return f(*[intermediate.enclosure for intermediate in args])
    return g

  primitive_to_enclosure_fun = {
      jax.lax.add_p: wrap(arithmetic.add),
      jax.lax.div_p: wrap(arithmetic.divide),
      jax.lax.integer_pow_p: pushforward_integer_pow,
      jax.lax.mul_p: wrap(arithmetic.multiply),
      jax.lax.neg_p: wrap(arithmetic.negative),
      jax.lax.pow_p: pushforward_pow,
      jax.lax.sub_p: wrap(arithmetic.subtract),
      # TODO(mstreeter): handle all bilinear primitives in a uniform way.
      jax.lax.dot_general_p: _dot_general_pushforward_fun(arithmetic),
      jax.lax.conv_general_dilated_p: _conv_general_dilated_pushforward_fun(
          arithmetic),
      jax.lax.broadcast_in_dim_p: _broadcast_in_dim_pushforward_fun,
  }
  primitive_to_enclosure_fun.update({
      primitive: _elementwise_pushforward_fun(arithmetic, get_enclosure)
      for primitive, get_enclosure in _ELEMENTWISE_PRIMITIVE_ENCLOSURES.items()
  })
  primitive_to_enclosure_fun.update({
      primitive: _pass_thru_pushforward_fun(primitive)
      for primitive in _PASS_THRU_PRIMITIVES
  })
  return primitive_to_enclosure_fun


def _validate_taylor_enclosure(a: types.TaylorEnclosureLike, x_shape):
  set_arithmetic = interval_arithmetic.IntervalArithmetic(jnp)
  for i, coeff in enumerate(a):
    s = set_arithmetic.shape(coeff)
    if s[len(s)-i*len(x_shape):] != i*x_shape:
      raise ValueError(x_shape, i, s, coeff, a)



# Copyright 2023 The autobound Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base class for unit tests."""

import math
from typing import List

from absl.testing import absltest
from autobound import types
import numpy as np


MAX_SIGMOID_DERIV = .25
MIN_SIGMOID_SECOND_DERIV = -0.09622504486493762
MAX_SIGMOID_SECOND_DERIV = 0.09622504486493762
MIN_SIGMOID_THIRD_DERIV = -0.125
MAX_SIGMOID_THIRD_DERIV = 0.04166666666666668


def sigmoid(x: float) -> float:
  return 1/(1+math.exp(-x)) if x >= 0 else math.exp(x)/(1+math.exp(x))


def sigmoid_derivative(order: int, x: float) -> float:
  """Returns order `order` derivative of sigmoid at `x`."""
  if order == 0:
    return sigmoid(x)
  elif order == 1:
    return sigmoid(x)*sigmoid(-x)
  elif order == 2:
    s = sigmoid(x)
    return s*sigmoid(-x)*(1-2*s)
  elif order == 3:
    s = sigmoid(x)
    sm = sigmoid(-x)
    return s*sm*((1-2*s)**2 - 2*s*sm)
  elif order == 4:
    s = sigmoid(x)
    sm = sigmoid(-x)
    return (s*sm*(1-2*s)*((1-2*s)**2 - 2*s*sm) +
            s*sm*(-4*(1-2*s)*s*sm - 2*s*sm*(1-2*s)))
  else:
    raise NotImplementedError(order)


def softplus(x: float) -> float:
  # Avoid overflow for large positive x using:
  # log(1+exp(x)) == log(1+exp(-|x|)) + max(x, 0).
  return math.log1p(math.exp(-abs(x))) + max(x, 0.)


def softplus_derivative(order: int, x: float) -> float:
  if order == 0:
    return softplus(x)
  else:
    return sigmoid_derivative(order-1, x)


def swish(x: float) -> float:
  return x*sigmoid(x)


def swish_derivative(order: int, x: float) -> float:
  return (order*sigmoid_derivative(order - 1, x) +
          x*sigmoid_derivative(order, x))


class TestCase(absltest.TestCase):
  """Base class for test cases."""

  def assert_allclose_strict(self, expected, actual, **kwargs):
    """Like np.testing.assert_allclose, but requires same shape/dtype."""
    np.testing.assert_allclose(actual, expected, **kwargs)
    self.assertEqual(np.asarray(expected).shape, np.asarray(actual).shape,
                     (expected, actual))
    self.assertEqual(np.asarray(expected).dtype, np.array(actual).dtype,
                     (expected, actual))

  def assert_enclosure_equal(self, expected, actual, **kwargs):
    self.assertLen(actual, len(expected))
    for e, a in zip(expected, actual):
      self.assert_interval_equal(e, a, **kwargs)

  def assert_interval_equal(self, expected, actual, **kwargs):
    if isinstance(expected, tuple):
      e0, e1 = expected
      self.assertIsInstance(actual, tuple)
      self.assertLen(actual, 2)
      a0, a1 = actual
      self.assert_allclose_strict(e0, a0, **kwargs)
      self.assert_allclose_strict(e1, a1, **kwargs)
    else:
      self.assert_allclose_strict(expected, actual, **kwargs)

  @classmethod
  def setUpClass(cls):
    super().setUpClass()
    cls.backends = _get_backends()


def _get_backends() -> List[types.NumpyLike]:
  """Returns list of NumpyLike back ends to test."""
  backends = [np]

  try:
    from jax.config import config as jax_config
    import jax.numpy as jnp
    backends.append(jnp)
    jax_config.update('jax_enable_x64', True)
  except ModuleNotFoundError:
    pass

  try:
    import tensorflow.experimental.numpy as tnp
    tnp.experimental_enable_numpy_behavior()
    backends.append(tnp)
  except ModuleNotFoundError:
    pass

  return backends